import java.awt.Color;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.image.BufferedImage;

class Dashboard {
	private static final int PF = GameView.PF;
	// dashboard variables
	private static int mDamageLineY = 0;
	private static int mDamageLineX = 0;
	/** Particles and clouds */
	private static final Particle[] mParticles = new Particle[32];
	private static boolean mIsSparksAnimation = false;
	/** false - start sparks animation, true - animation finished */
	private static boolean mIsSparksAnimationFinished = true;
	/** false - start cloud animation, true - animation finished */
	private static boolean mIsCloudAnimationFinished = true;
	/** time to show lap time. 0 - to show current time, > 0 - to freeze lap time */
	private static int mShowLapTimeTimer = 0;// int16
	/** true, if lap time was improved */
	private static boolean mIsLapTimeImproved = false;
	/** timer to update car to car distance indicator */
	private static int mUpdateDistanceTimer = 0;
	/** race time in %1m:%2ss.%2ms */
	static final char[] mStrLapTime = { ' ', ':', ' ', ' ', '.', ' ', ' ' };
	/** best lap time in %1m:%2ss.%2ms */
	private static final char[] mStrBestLapTime = { ' ', ':', ' ', ' ', '.', ' ', ' ' };
	/** car to car distance */
	private static final char[] mStrDistance = { ' ', ' ', ' ', ' ', ' ' };
	private static final char[] mStrLap = { 'L', ' ' };
	private static final char[] mStrBoost = { 'B', ' ', ' ', ' ' };
	// end dashboard variables
	private static Image imgLightWatch = null;
	private static Image[] imgClouds = new Image[7];
	private static Image imgWatch = null;
	private static int mWheelFrameNumber = 0; // int8
	private static Image imgFlag = null;
	private static Image imgRightCorner = null;
	// images
	private static Image imgCockpit = null;
	private static int nFlameNumber = 0; // int8
	private static Image imgLeftPillar = null;
	private static int mWheelSpeed = 0; // int16
	private static int mCloudFrameNumber = 0; // int8
	private static Image[] imgRightWheel = new Image[3];
	private static Image imgTop = null;
	private static Image imgRightPillar = null;
	/** 3 frames for 4 left exhaust pipe */
	private static Image[] imgLeftFlames = new Image[3 * 4];
	private static Image imgLeftCorner = null;
	/** 3 frames for 4 right exhaust pipe */
	private static Image[] imgRightFlames = new Image[3 * 4];
	private static int mWheelRunDistance = 0; // int16
	private static Image[] imgLeftWheel = new Image[3];
	private static Image imgSteer = null;
	private static Image imgChain = null;
	private static int mFlameFrameNumber = 0; // int8
	//
	static int init() {
		//
		for( int i = 0; i < mParticles.length; i++ ) {
			final Particle p = new Particle();
			p.mY_fp8 = GameView.SCREEN_HEIGHT << 8;// off screen
			mParticles[i] = p;
		}
		// load images
		try {
			imgCockpit = GameView.getImage("db").getScaledInstance( 320 * PF, 77 * PF, BufferedImage.SCALE_SMOOTH );
			//
			imgSteer = GameView.getImage("steer").getScaledInstance( 10 * PF, 5 * PF, BufferedImage.SCALE_SMOOTH );
			//
			imgFlag = GameView.getImage("flag").getScaledInstance( 12 * PF, 8 * PF, BufferedImage.SCALE_SMOOTH );
			//
			imgTop = GameView.getImage("top").getScaledInstance( 24 * PF, 16 * PF, BufferedImage.SCALE_SMOOTH );
			//
			BufferedImage bimg = GameView.getImage("watch");
			imgWatch = bimg.getSubimage( 0, 0, 12, 7 )
							.getScaledInstance( 12 * PF, 7 * PF, BufferedImage.SCALE_SMOOTH );
			imgLightWatch = bimg.getSubimage( 12, 0, 12, 7 )
							.getScaledInstance( 12 * PF, 7 * PF, BufferedImage.SCALE_SMOOTH );
			//
			bimg = GameView.getImage("pillar");
			imgLeftPillar = bimg.getSubimage( 0, 0, 16, 101 )
							.getScaledInstance( 16 * PF, 101 * PF, BufferedImage.SCALE_SMOOTH );
			imgRightPillar = bimg.getSubimage( 16, 0, 16, 101 )
							.getScaledInstance( 16 * PF, 101 * PF, BufferedImage.SCALE_SMOOTH );
			//
			bimg = GameView.getImage("corner");
			imgLeftCorner = bimg.getSubimage( 0, 0, 22, 22 )
							.getScaledInstance( 22 * PF, 22 * PF, BufferedImage.SCALE_SMOOTH );
			imgRightCorner = bimg.getSubimage( 22, 0, 22, 22 )
							.getScaledInstance( 22 * PF, 22 * PF, BufferedImage.SCALE_SMOOTH );
			// wheels
			bimg = GameView.getImage("w");
			for( int i = 0; i < imgLeftWheel.length; i++ ) {
				imgLeftWheel[i] = bimg.getSubimage( i * 24, 0, 24, 55 )
						.getScaledInstance( 24 * PF, 55 * PF, BufferedImage.SCALE_SMOOTH );
			}
			for( int i = 0; i < imgRightWheel.length; i++ ) {
				imgRightWheel[i] = bimg.getSubimage( 144 - (i + 1) * 24, 0, 24, 55 )
						.getScaledInstance( 24 * PF, 55 * PF, BufferedImage.SCALE_SMOOTH );
			}
			// chain
			imgChain = GameView.getImage("chain").getScaledInstance( 10 * PF, 16 * PF, BufferedImage.SCALE_SMOOTH );
			// flames
			bimg = GameView.getImage("f");
			for( int i = 0; i < (imgLeftFlames.length >> 2); i++ ) {
				imgLeftFlames[0 * 3 + i] = bimg.getSubimage( i * 15, 0, 15, 20 )
						.getScaledInstance( 15 * PF, 20 * PF, BufferedImage.SCALE_SMOOTH );
				imgRightFlames[0 * 3 + i] = bimg.getSubimage( 90 - (i + 1) * 15, 0, 15, 20 )
						.getScaledInstance( 15 * PF, 20 * PF, BufferedImage.SCALE_SMOOTH );
				imgLeftFlames[1 * 3 + i] = bimg.getSubimage( i * 15, 20, 15, 14 )
						.getScaledInstance( 15 * PF, 14 * PF, BufferedImage.SCALE_SMOOTH );
				imgRightFlames[1 * 3 + i] = bimg.getSubimage( 90 - (i + 1) * 15, 20, 15, 14 )
						.getScaledInstance( 15 * PF, 14 * PF, BufferedImage.SCALE_SMOOTH );
				imgLeftFlames[2 * 3 + i] = bimg.getSubimage( i * 14, 34, 14, 14 )
						.getScaledInstance( 14 * PF, 14 * PF, BufferedImage.SCALE_SMOOTH );
				imgRightFlames[2 * 3 + i] = bimg.getSubimage( 90 - (i + 1) * 14, 34, 14, 14 )
						.getScaledInstance( 14 * PF, 14 * PF, BufferedImage.SCALE_SMOOTH );
				imgLeftFlames[3 * 3 + i] = bimg.getSubimage( i * 9, 48, 9, 10 )
						.getScaledInstance( 9 * PF, 10 * PF, BufferedImage.SCALE_SMOOTH );
				imgRightFlames[3 * 3 + i] = bimg.getSubimage( 90 - (i + 1) * 9, 48, 9, 10 )
						.getScaledInstance( 9 * PF, 10 * PF, BufferedImage.SCALE_SMOOTH );
			}
			bimg = GameView.getImage("clouds");
			for( int i = 0; i < imgClouds.length; i++ ) {
				imgClouds[i] = bimg.getSubimage( 0, i * 39, 65, 39 )
						.getScaledInstance( 65 * PF, 39 * PF, BufferedImage.SCALE_SMOOTH );
			}
		}
		catch(Exception ie) {
			System.out.println( ie.getMessage() );
			return -1;
		}
		return 0;
	}
	static void initRace() {
		mIsSparksAnimationFinished = true;
		mIsCloudAnimationFinished = true;
		// init damage line
		mDamageLineY = 2 * PF;// 2 is vertical margin
		mDamageLineX = 22 * PF;// 22 is horizontal margin, edge pillar width
		// init car to car distance
		mStrDistance[0] = ' ';
		for( int i = 1; i < mStrDistance.length; i++ ) mStrDistance[i] = '0';
		// lap time
		mShowLapTimeTimer = 0;
		mStrLapTime[1] = ':';
		mStrLapTime[4] = '.';
		for( int i = 0; i < mStrBestLapTime.length; i++ ) mStrBestLapTime[i] = ' ';
		//for( int i = 0; i < mStrRaceTime.length; i++ ) mStrRaceTime[i] = ' ';
		mWheelSpeed = 0;
	}
	static void stopAnimation() {
		mWheelSpeed = 0;
		mIsCloudAnimationFinished = true;
		mIsSparksAnimationFinished = true;
	}
	static void freezeLapTime(boolean isImproved) {
		mShowLapTimeTimer = (3 * 1000 / (int)GameView.REFRESH_PERIOD);// 3 sec to show lap time
		if( isImproved ) {//copy best lap time
			System.arraycopy( mStrLapTime, 0, mStrBestLapTime, 0, mStrBestLapTime.length );
		}
		mIsLapTimeImproved = isImproved;
	}
	/** Render car pillars and set clip area */
	static void renderPillars(final Graphics g) {
		final int cockpit_height = 77 * PF;
		final int pillar_width = 16 * PF;
		final int top_pillar_width = 24 * PF;
		final int top_pillar_height = 16 * PF;
		final int edge_pillar_width = 22 * PF;
		final int edge_pillar_height = 22 * PF;
		//
		g.drawImage( imgCockpit, 0, GameView.SCREEN_HEIGHT - cockpit_height, null );
		g.drawImage( imgLeftPillar, 0, edge_pillar_height, null );
		g.drawImage( imgRightPillar, GameView.SCREEN_WIDTH - pillar_width, edge_pillar_height, null );
		final int right = GameView.SCREEN_WIDTH - edge_pillar_width;
		for( int h = edge_pillar_width; h < right; h += top_pillar_width ) {
			g.drawImage( imgTop, h, 0, null );
		}
		g.drawImage( imgLeftCorner, 0, 0, null );
		g.drawImage( imgRightCorner, GameView.SCREEN_WIDTH - edge_pillar_width, 0, null );
		// set clip to prevent drawing on pillars
		g.setClip( pillar_width, top_pillar_height, GameView.SCREEN_WIDTH - pillar_width * 2, GameView.SCREEN_HEIGHT - top_pillar_height );
	}
	private static void renderSparks(final Graphics g, int speed, boolean isContinue) {
		final int cockpit_height = 77 * PF;
		//
		g.setColor( Color.YELLOW );
		speed = (4096 << 8) + (speed << (8 + 4));
		final int visible_y_fp8 = (GameView.SCREEN_HEIGHT - cockpit_height + 22 * PF) << 8;
		boolean is_finished = true;
		for( int i = 0; i < mParticles.length; i++ ) {
			final Particle p = mParticles[i];
			if( isContinue && (p.mY_fp8 > visible_y_fp8 || p.mX_fp8 < 0 || p.mX_fp8 > (GameView.SCREEN_WIDTH << 8)) ) {
				p.mY_fp8 = visible_y_fp8;
				p.mX_fp8 = GameView.SCREEN_HALFWIDTH << 8;
				int acceleration_fp8 = speed;
				acceleration_fp8 /= (int)(3 * 1000 * 1000 / GameView.REFRESH_PERIOD / GameView.REFRESH_PERIOD / PF);
				acceleration_fp8 += GameView.mRandom.nextInt( acceleration_fp8 );
				p.mDY_fp8 -= acceleration_fp8;
				p.mDX_fp8 = (GameView.mRandom.nextInt( GameView.SCREEN_WIDTH ) - GameView.SCREEN_HALFWIDTH) << 4;
			}
			p.mY_fp8 += p.mDY_fp8;
			if( p.mY_fp8 <= visible_y_fp8 ) is_finished = false;
			// gravitation, 1.7 m/s
			p.mDY_fp8 += (int)(256 * 170 * PF * GameView.REFRESH_PERIOD * GameView.REFRESH_PERIOD / 1000 / 1000);
			p.mX_fp8 += p.mDX_fp8;
			g.fillRect( p.mX_fp8 >> 8, p.mY_fp8 >> 8, 2 * PF, 2 * PF );
		}
		mIsSparksAnimationFinished = is_finished;
	}
	static void initClouds() {
		if( mIsCloudAnimationFinished ) {// initiate clouds
			mIsCloudAnimationFinished = false;
			SoundManager.play( SoundManager.OFF_ROAD, 256, 256 );
			final int cockpit_height = 77 * PF;
			final int cloud_width = 65 * PF;
			//
			final int start_y_fp8 = (GameView.SCREEN_HEIGHT - cockpit_height + 22 * PF) << 8;
			for( int i = 0; i < mParticles.length; i++ ) {
				final Particle p = mParticles[i];
				p.mFrame = GameView.mRandom.nextInt( 7 );
				p.mY_fp8 = start_y_fp8;
				p.mX_fp8 = GameView.mRandom.nextInt( GameView.SCREEN_WIDTH - cloud_width ) << 8;
				// 8.5 m/s
				final int acceleration_fp8 = (int)(256 * 850 * PF * GameView.REFRESH_PERIOD * GameView.REFRESH_PERIOD / 1000 / 1000);
				p.mDY_fp8 = -(acceleration_fp8 + GameView.mRandom.nextInt( acceleration_fp8 ));
				p.mDX_fp8 = (GameView.mRandom.nextInt( GameView.SCREEN_WIDTH ) - GameView.SCREEN_HALFWIDTH) << 2;
			}
		}
	}
	static void renderClouds(final Graphics g) {
		if( mIsCloudAnimationFinished ) {
			return;
		}
		final int cockpit_height = 77 * PF;
		//
		int frame_shift = 0;
		if( ++mCloudFrameNumber >= (int)(1000 / GameView.REFRESH_PERIOD / 8) ) {
			mCloudFrameNumber = 0;
			frame_shift = 1;//mRandom.nextInt( 7 );
		}
		boolean is_finished = true;
		final int visible_y_fp8 = (GameView.SCREEN_HEIGHT - cockpit_height + 22 * PF) << 8;
		for( int i = 0; i < mParticles.length; i++ ) {
			final Particle p = mParticles[i];
			if( p.mY_fp8 <= visible_y_fp8 ) is_finished = false;
			p.mY_fp8 += p.mDY_fp8;
			// gravitation, 0.66 m/s
			p.mDY_fp8 += (int)(256 * 66 * PF * GameView.REFRESH_PERIOD * GameView.REFRESH_PERIOD / 1000 / 1000);
			p.mX_fp8 += p.mDX_fp8;
			//
			p.mFrame += frame_shift;
			if( p.mFrame >= 7 ) p.mFrame -= 7;
			g.drawImage( imgClouds[p.mFrame], p.mX_fp8 >> 8, p.mY_fp8 >> 8, null );
		}
		mIsCloudAnimationFinished = is_finished;
	}
	private static void renderChains(final Graphics g, int phase) {
		final int cockpit_height = 77 * PF;
		final int chain_width = 10 * PF;
		final int chain_pos = 55 * PF;
		final int chain_height = 16 * PF;
		final int chain_right_x = GameView.SCREEN_WIDTH - (chain_pos + chain_width);
		final int bottom = GameView.SCREEN_HEIGHT - cockpit_height + chain_width;
		final int step = (bottom - 0) / PlayerCar.CHAIN_RELEASE_TIME;
		//
		for( int y = bottom - step * phase; y > 0; y -= chain_height ) {
			g.drawImage( imgChain, chain_pos, y, null );// left
			g.drawImage( imgChain, chain_right_x, y, null );// right
		}
	}
	private static void renderDamage(final Graphics g, int damage) {
		final int top_pillar_height = 16 * PF;
		final int pillar_width = 16 * PF;
		final int damage_line_margin = 22;// not scaled, edge pillar width
		final int damage_line_pos_y = 2 * PF;
		final int damage_line_deviation = 3 * PF;
		//
		g.setClip( 0, 0,GameView. SCREEN_WIDTH, top_pillar_height );
		//
		// draw damage line
		//g.setColor( Color.RED );
		//final int damage_line_length = damage * (SCREEN_WIDTH - damage_line_margin * 2) / Car.DAMAGE_LIMIT;
		//g.fillRect( damage_line_margin, 2 * PF, damage_line_length, 4 * PF );
		final int damage_line_end = (damage_line_margin + damage * (GameView.SCREEN_WIDTH / PF - damage_line_margin * 2) / Car.DAMAGE_LIMIT) * PF;
		//
		for( int i = mDamageLineX; i < damage_line_end; i += 1 * PF ) {
			int direction = GameView.mRandom.nextInt( 3 );// 0 - down, 1 - up, 2 - straight
			Color color = new Color( 128, 64, 64 );
			if( direction == 0 ) {
				if( mDamageLineY < (damage_line_pos_y + damage_line_deviation - 1 * PF) ) {
					mDamageLineY += PF;
					color = new Color( 192, 128, 128 );// to mark falling
				}
				//if( mDamageLineY < (damage_line_pos_y + damage_line_deviation - 1) ) mDamageLineY++;
			}
			if( --direction == 0 ) {
				if( mDamageLineY > damage_line_pos_y ) mDamageLineY -= PF;
			}
			// change color to mark increment
			//g.setColor( i == mDamageLineX ? new Color( 192, 128, 128 ) : new Color( 128, 64, 64 ) );
			g.setColor( color );
			g.fillRect( i, mDamageLineY, 1 * PF, 1 * PF );
			g.setColor( Color.BLACK );
			g.fillRect( i, mDamageLineY + PF, 1 * PF, 2 * PF );
		}
		mDamageLineX = damage_line_end;
		//
		// draw opponent damage line to test
		//g.setColor( Color.BLUE );
		//final int opp_damage_line_length = mOpponentCar.mDamage * (SCREEN_WIDTH - damage_line_margin * 2) / Car.DAMAGE_LIMIT;
		//g.fillRect( damage_line_margin, (2 + 4) * PF, opp_damage_line_length, 4 * PF );
		// draw damage count
		//g.setColor( Color.BLACK );
		//g.drawString( Integer.toString( damage ), damage_line_margin, 2 * PF + FONT_BASELINE );
		//
		g.setClip( pillar_width, top_pillar_height, GameView.SCREEN_WIDTH - pillar_width * 2, GameView.SCREEN_HEIGHT - top_pillar_height );
	}
	private static void calculateWheelFrameNumber(final Car car) {
		if( car.mIsTouchingRoad ) {
			mWheelSpeed = Math.abs( car.mSpeed_fp8 );
		} else {
			mWheelSpeed -= mWheelSpeed >> 4;
		}
		mWheelRunDistance += mWheelSpeed;
		if( mWheelRunDistance > (int)(256 * 1000 / GameView.REFRESH_PERIOD) ) {
			mWheelRunDistance = 0;
			if( car.mSpeed_fp8 > 0 ) {
				if( ++mWheelFrameNumber >= imgLeftWheel.length ) mWheelFrameNumber = 0;
				return;
			}
			if( --mWheelFrameNumber < 0 ) mWheelFrameNumber += imgLeftWheel.length;
		}
	}
	/**
	 * @param ms10 - time in ms / 10
	 * @return filled mStrLapTime[0,2,3,5,6]
	 */
	static void convertMsToChars(final int ms10) {
		if( ms10 >= 0x00ffffff ) {
			for( int i = 0; i < mStrLapTime.length; i++ ) mStrLapTime[i] = '-';
			return;
		}
		final int min = Math.min( ms10 / 6000, 9 );
		final int sec = (ms10 / 100) % 60;
		mStrLapTime[0] = (char)(min + '0');
		mStrLapTime[2] = (char)(sec / 10 + '0');
		mStrLapTime[3] = (char)(sec % 10 + '0');
		mStrLapTime[5] = (char)(((ms10 / 10) % 10) + '0');
		mStrLapTime[6] = (char)((ms10 % 10) + '0');
	}
	/* test 3d projection
	private void renderFrontWheelScreenPositions(final Graphics g, final Car car) {
		// -CAR_WIDTH/2, LOOK_FRONT_WHEEL, LOOK_HEIGHT - WHEEL_RADIUS
		// CAR_WIDTH/2, LOOK_FRONT_WHEEL, LOOK_HEIGHT - WHEEL_RADIUS
		final int wheelHalfWidth = 2 * PF;
		int scale = (Render.SCALE << Camera.PRECISION_LOG2) / (Car.WHEEL_FRONT_Y + Render.CAMERA_FOV);
		int x = (scale * (Car.WHEEL_RIGHT_X)) >> Camera.PRECISION_LOG2;
		// draw centers
		final int wheel_center_y = Car.WHEEL_BOTTOM_Z + Car.WHEEL_RADIUS + Car.SPRINT_TRAVEL / 2 - Car.LOOK_Z;
		//int y = mLookY - ((scale * (car.mFrontLeftWheel_BottomZ + car.WHEEL_RADIUS - car.LOOK_Z)) >> PRECISION_LOG2);
		int y = Render.LOOK_Y - ((scale * (wheel_center_y)) >> Camera.PRECISION_LOG2);
		g.setColor( Color.BLACK );
		g.fillRect( GameView.SCREEN_HALFWIDTH - x - wheelHalfWidth, y - wheelHalfWidth, wheelHalfWidth*2, wheelHalfWidth*2 );
		g.fillRect( GameView.SCREEN_HALFWIDTH + x - wheelHalfWidth, y - wheelHalfWidth, wheelHalfWidth*2, wheelHalfWidth*2 );
		// draw bottom
		//y = mLookY - ((scale * (car.mFrontLeftWheel_BottomZ - car.LOOK_Z)) >> PRECISION_LOG2);
		y = Render.LOOK_Y - ((scale * (wheel_center_y - Car.WHEEL_RADIUS)) >> Camera.PRECISION_LOG2);
		g.fillRect( GameView.SCREEN_HALFWIDTH - x - wheelHalfWidth, y - wheelHalfWidth, wheelHalfWidth*2, wheelHalfWidth*2 );
		g.fillRect( GameView.SCREEN_HALFWIDTH + x - wheelHalfWidth, y - wheelHalfWidth, wheelHalfWidth*2, wheelHalfWidth*2 );
		// draw top
		//y = mLookY - ((scale * (car.mFrontLeftWheel_BottomZ + 2 * car.WHEEL_RADIUS - car.LOOK_Z)) >> PRECISION_LOG2);
		y = Render.LOOK_Y - ((scale * (wheel_center_y + Car.WHEEL_RADIUS)) >> Camera.PRECISION_LOG2);
		g.fillRect( GameView.SCREEN_HALFWIDTH - x - wheelHalfWidth, y - wheelHalfWidth, wheelHalfWidth*2, wheelHalfWidth*2 );
		g.fillRect( GameView.SCREEN_HALFWIDTH + x - wheelHalfWidth, y - wheelHalfWidth, wheelHalfWidth*2, wheelHalfWidth*2 );
		// front
		scale = (Render.SCALE << Camera.PRECISION_LOG2) / (Car.WHEEL_FRONT_Y + Car.WHEEL_RADIUS + Render.CAMERA_FOV);
		x = (scale * (Car.WIDTH / 2)) >> Camera.PRECISION_LOG2;
		//y = mLookY - ((scale * (car.mFrontLeftWheel_BottomZ + car.WHEEL_RADIUS - car.LOOK_Z)) >> PRECISION_LOG2);
		y = Render.LOOK_Y - ((scale * (wheel_center_y)) >> Camera.PRECISION_LOG2);
		g.fillRect( GameView.SCREEN_HALFWIDTH - x - wheelHalfWidth, y - wheelHalfWidth, wheelHalfWidth*2, wheelHalfWidth*2 );
		g.fillRect( GameView.SCREEN_HALFWIDTH + x - wheelHalfWidth, y - wheelHalfWidth, wheelHalfWidth*2, wheelHalfWidth*2 );
		// rear
		scale = (Render.SCALE << Camera.PRECISION_LOG2) / (Car.WHEEL_FRONT_Y - Car.WHEEL_RADIUS + Render.CAMERA_FOV);
		x = (scale * (Car.WIDTH / 2)) >> Camera.PRECISION_LOG2;
		//y = mLookY - ((scale * (car.mFrontLeftWheel_BottomZ + car.WHEEL_RADIUS - car.LOOK_Z)) >> PRECISION_LOG2);
		y = Render.LOOK_Y - ((scale * (wheel_center_y)) >> Camera.PRECISION_LOG2);
		g.fillRect( GameView.SCREEN_HALFWIDTH - x - wheelHalfWidth, y - wheelHalfWidth, wheelHalfWidth*2, wheelHalfWidth*2 );
		g.fillRect( GameView.SCREEN_HALFWIDTH + x - wheelHalfWidth, y - wheelHalfWidth, wheelHalfWidth*2, wheelHalfWidth*2 );
		// draw car body
		int dz = Car.WHEEL_RADIUS + 32 - (Car.CG_Z + Car.LOOK_Z);
		scale = (Render.SCALE << Camera.PRECISION_LOG2) / (Car.BONNET + Render.CAMERA_FOV);
		x = (scale * Car.WIDTH / 2) >> Camera.PRECISION_LOG2;
		y = (scale * dz) >> Camera.PRECISION_LOG2;
		g.drawLine( GameView.SCREEN_HALFWIDTH - x, Render.LOOK_Y - y, GameView.SCREEN_HALFWIDTH + x, Render.LOOK_Y - y );
		// draw panel
		dz = -16;// panel
		int dy = 45 + Car.LENGTH / 2 - Car.WHEEL_RADIUS;
		int dx = 20;
		scale = (Render.SCALE << Camera.PRECISION_LOG2) / (dy + Render.CAMERA_FOV);
		x = (scale * dx) >> Camera.PRECISION_LOG2;
		y = (scale * dz) >> Camera.PRECISION_LOG2;
		//g.setColor( new Color(115, 48, 49) );
		g.setColor( new Color(49, 85, 115) );
		//g.fillRect( mScreenHalfWidth - x, mScreenHalfHeight - y, x << 1, mScreenHeight + y - mScreenHalfHeight );
		g.drawRect( GameView.SCREEN_HALFWIDTH - x, Render.LOOK_Y - y, x << 1, GameView.SCREEN_HEIGHT + y - Render.LOOK_Y );
		// draw steering wheel
		final int radius = 15;
		dy = 30 + Car.LENGTH / 2 - Car.WHEEL_RADIUS;
		dz = -15 - radius;
		scale = (Render.SCALE << Camera.PRECISION_LOG2) / (dy + Render.CAMERA_FOV);
		int r = (scale * radius) >> Camera.PRECISION_LOG2;
		y = (scale * dz) >> Camera.PRECISION_LOG2;
		g.setColor( Color.BLACK );
		g.drawOval( GameView.SCREEN_HALFWIDTH - r, Render.LOOK_Y - y, r << 1, r << 1 );
	}
	*/
	static void render(final Graphics g, final PlayerCar car, final int keys, final boolean isOpponentBehind) {
		calculateWheelFrameNumber( car );
		int speed = Math.max( 0, car.getSpeedKmH() );
		if( car.mIsOnRoad && (car.mIsScratchSide || (car.mDamage >= Car.DAMAGE_LIMIT)) ) {
			mIsSparksAnimation = true;
			mIsSparksAnimationFinished = false;
		}
		if( ! mIsSparksAnimationFinished ) {
			renderSparks( g, speed, mIsSparksAnimation );
			mIsSparksAnimation = false;
		}
		if( car.mChainReleaseCounter < PlayerCar.CHAIN_RELEASE_TIME ) {
			if( ! car.mIsOnChains ) ++car.mChainReleaseCounter;
			renderChains( g, car.mChainReleaseCounter );
		}
		/* draw viewport
		g.setColor( new Color( 0x60, 0x20, 0x20 ) );
		g.drawRoundRect( (mScreenWidth - mViewportWidth) >> 1,
				mLookY - (mViewportHeight >> 1),
				mViewportWidth, mViewportHeight, 6 * PF, 6 * PF );
		*/
		// draw position and vector
		/*
		final int xcenter = 22 * PF;
		final int ycenter = (60 + 22) * PF;
		// draw track
		g.setColor( Color.WHITE );
		for( int i = 0; i < mTrackPoints.length; i += 2 ) {
			g.fillRect( mTrackPoints[i], mTrackPoints[i + 1], 1, 1 );
		}
		*/
		// draw marker. 8px/10kmh. starts from 10 kmh.
		final int indicator_length = Math.max( Math.min( speed * (PF * 8) / 10, 155 * PF ) - 10 * PF * 8 / 10, 0 );
		/*
		final int scale = 600 / PF;
		final int indicator_width = 3 * PF;
		g.setColor( Color.RED );
		int x = xcenter + car.mX / scale;
		int y = ycenter - car.mY / scale;
		int xend = x + ((indicator_length * car.sin_z) >> (Camera.PRECISION_LOG2 + 3));
		int yend = y - ((indicator_length * car.cos_z) >> (Camera.PRECISION_LOG2 + 3));
		final int dx = ((indicator_width * car.cos_z) >> Camera.PRECISION_LOG2);
		final int x1 = x + dx;
		final int x2 = x - dx;
		final int dy = ((indicator_width * car.sin_z) >> Camera.PRECISION_LOG2);
		final int y1 = y + dy;
		final int y2 = y - dy;
		g.drawLine( x1, y1, xend, yend );
		g.drawLine( x2, y2, xend, yend );
		g.drawLine( x1, y1, x2, y2 );
		// draw opponent car position
		g.setColor( Color.RED );
		x = xcenter + (mOpponentCar.mCG_X) / scale;
		y = ycenter - (mOpponentCar.mCG_Y) / scale;
		g.fillRect( x, y, 1 * PF, 1 * PF );
		*/
		// calculate cockpit y position
		//final int s = (Render.SCALE << Camera.PRECISION_LOG2) / (car.BONNET + Render.CAMERA_FOV);
		final int cockpit_width = 320 * PF;
		final int cockpit_height = 77 * PF;
		final int cockpit_x = (GameView.SCREEN_WIDTH - cockpit_width) >> 1;
		final int cockpit_right_x = cockpit_x + cockpit_width;
		//final int dz = -20;
		final int cockpit_y = GameView.SCREEN_HEIGHT - cockpit_height;//Render.LOOK_Y - ((s * dz) >> Camera.PRECISION_LOG2);
		// draw wheels
		final int wscale = (Render.SCALE << Camera.PRECISION_LOG2) / (Car.WHEEL_FRONT_Y + Render.CAMERA_FOV);
		final int wy_right = Render.LOOK_Y - (wscale * (car.mFrontRightWheelBottom + (Car.WHEEL_RADIUS << 1) - Car.LOOK_Z) >> Camera.PRECISION_LOG2);
		final int wy_left = Render.LOOK_Y - (wscale * (car.mFrontLeftWheelBottom + (Car.WHEEL_RADIUS << 1) - Car.LOOK_Z) >> Camera.PRECISION_LOG2);
		g.drawImage( imgLeftWheel[mWheelFrameNumber], cockpit_x + 16 * PF, wy_left, null );
		g.drawImage( imgRightWheel[mWheelFrameNumber], cockpit_x + cockpit_width - (16 + 24) * PF, wy_right, null );
		// draw cockpit
		g.drawImage( imgCockpit, cockpit_x, cockpit_y, null );
		// draw corners
		g.drawImage( imgLeftCorner, 0, 0, null );
		g.drawImage( imgRightCorner, GameView.SCREEN_WIDTH - 22 * PF, 0, null );
		// draw flame
		if( car.mIsBoostActivated ) {
			if( ++mFlameFrameNumber >= (int)(1000 / GameView.REFRESH_PERIOD / 8) ) {
				mFlameFrameNumber = 0;
				if( ++nFlameNumber >= 3 ) nFlameNumber = 0;
			}
			g.drawImage( imgLeftFlames[0 * 3 + nFlameNumber], cockpit_x + 32 * PF, cockpit_y + 8 * PF, null );
			g.drawImage( imgLeftFlames[1 * 3 + nFlameNumber], cockpit_x + 47 * PF, cockpit_y + 7 * PF, null );
			g.drawImage( imgLeftFlames[2 * 3 + nFlameNumber], cockpit_x + 58 * PF, cockpit_y + 1 * PF, null );
			g.drawImage( imgLeftFlames[3 * 3 + nFlameNumber], cockpit_x + 71 * PF, cockpit_y + 0 * PF, null );
			g.drawImage( imgRightFlames[0 * 3 + nFlameNumber], cockpit_right_x - (32 + 15) * PF, cockpit_y + 8 * PF, null );
			g.drawImage( imgRightFlames[1 * 3 + nFlameNumber], cockpit_right_x - (47 + 15) * PF, cockpit_y + 7 * PF, null );
			g.drawImage( imgRightFlames[2 * 3 + nFlameNumber], cockpit_right_x - (58 + 14) * PF, cockpit_y + 1 * PF, null );
			g.drawImage( imgRightFlames[3 * 3 + nFlameNumber], cockpit_right_x - (71 + 9) * PF, cockpit_y + 0 * PF, null );
		}
		// draw speed indicator
		g.setColor( Color.YELLOW );
		g.fillRect( cockpit_x + 87 * PF, cockpit_y + 51 * PF, indicator_length, 2 * PF );
		// start draw digit information
		g.setColor( Color.BLACK );
		// draw lap number
		mStrLap[1] = (char)(Math.max( car.mLap, 0 ) + '0');
		g.drawChars( mStrLap, 0, 2, cockpit_x + 25 * PF, cockpit_y + 62 * PF );
		// draw boost reserve
		mStrBoost[1] = (char)(car.mBoostReserve / 100 + '0');
		mStrBoost[2] = (char)((car.mBoostReserve / 10) % 10 + '0');
		mStrBoost[3] = (char)(car.mBoostReserve % 10 + '0');
		g.drawChars( mStrBoost, 0, 4, cockpit_x + 45 * PF, cockpit_y + 62 * PF );
		// draw speed digit
		//g.drawString( Integer.toString( speed ), cockpit_x + 24 * PF, cockpit_y + 72 * PF );
		// draw distance between cars. If practice, have no car to car interaction
		if( /*! mIsPractice && */++mUpdateDistanceTimer >= (int)(1000 / GameView.REFRESH_PERIOD / 2) ) {
			mUpdateDistanceTimer = 0;
			int val = Car.sDistance;
			mStrDistance[0] = ' ';
			if( val < 0 ) {
				val = -val;
				mStrDistance[0] = '-';
			}
			mStrDistance[1] = (char)(val / 1000 + '0');
			mStrDistance[2] = (char)((val / 100) % 10 + '0');
			mStrDistance[3] = (char)((val / 10) % 10 + '0');
			mStrDistance[4] = (char)((val % 10) + '0');
		}
		g.drawChars( mStrDistance, 0, 5, cockpit_x + 40 * PF, cockpit_y + 72 * PF );
		// flag
		if( isOpponentBehind ) {
			g.drawImage( imgFlag, cockpit_x + 83 * PF, cockpit_y + 67 * PF, null );
		}
		// watch
		Image img = imgWatch;
		if( mIsLapTimeImproved )
			img = imgLightWatch;
		g.drawImage( img, cockpit_x + 225 * PF, cockpit_y + 68 * PF, null );
		// draw current time "min:sec.ms", 0:00.0
		if( car.mStartRaceTime != 0 ) {
			if( mShowLapTimeTimer <= 0 ) {
				convertMsToChars( (int)(System.currentTimeMillis() - car.mStartLapTime) / 10 );
			} else {
				mShowLapTimeTimer--;
			}
			g.drawChars( mStrLapTime, 0, 6, cockpit_x + 254 * PF, cockpit_y + 62 * PF );
			// draw best lap time "min:sec.ms", 0:00.00
			g.drawChars( mStrBestLapTime, 0, 7, cockpit_x + 254 * PF, cockpit_y + 72 * PF );
		}
		// draw steer
		final int steer_width = 10 * PF;
		final int steer_height = 5 * PF;
		if( (keys & (GameView.KEY_LEFT | GameView.KEY_RIGHT)) != 0 ) {
			final int steer_y = cockpit_y + cockpit_height - steer_height;
			g.drawImage( imgSteer, (GameView.SCREEN_WIDTH - steer_width) >> 1, steer_y, null );
		}
		// damage
		if( car.mIsUpdateDamage ) renderDamage( g, car.mDamage );
		// to test projection
		//renderFrontWheelScreenPositions( mG, mCar );
	}
}