import java.awt.AlphaComposite;
import java.awt.Color;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.GraphicsEnvironment;
import java.awt.Image;
import java.awt.Insets;
import java.awt.Rectangle;
import java.awt.RenderingHints;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.WindowEvent;
import java.awt.event.WindowListener;
import java.awt.image.BufferedImage;
import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.UnsupportedEncodingException;
import java.util.Properties;
import java.util.Random;

import javax.imageio.ImageIO;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.SwingUtilities;
import javax.swing.UIManager;

class GameView extends JFrame implements KeyListener, ComponentListener, Runnable {
	private static final long serialVersionUID = 2L;
	static final String APP_NAME = "StuntCarRacer";
	private static final String TEXT_ENCODE = "UTF-8";
	private Thread mThread = null;
	/** screen refresh period, ms. fps = 1000 / REFRESH_PERIOD<br>
	 * int64 to avoid incorrect calculating time dependent constants
	 */
	static final long REFRESH_PERIOD = 50;// ms, 20 Hz
	/** cross fade time, ms */
	private static final int CROSS_FADE_MS = 500;
	private volatile boolean mIsRunning = false;
	private final Object mPauseLock = new Object();
	private volatile boolean mIsPaused = false;
	private final Object mPaintLock = new Object();
	private Image mImgOffscreenBuffer = null;
	private Graphics mG = null;
	/** Picture factor */
	static final int PF = 3;
	static final int SCREEN_WIDTH = 320 * PF;
	static final int SCREEN_HEIGHT = 200 * PF;
	static final int SCREEN_HALFWIDTH = SCREEN_WIDTH / 2;
	static final int SCREEN_HALFHEIGHT = SCREEN_HEIGHT / 2;
	private int mFrameViewportTopMargin = 0;
	private int mFrameViewportLeftMargin = 0;
	//private final int mViewportHeight;
	//private final int mViewportWidth;
	/** Adopt to current screen width size */
	private double mPaintScaleWidth = 1.0;
	/** Adopt to current screen height size */
	private double mPaintScaleHeight = 1.0;
	/** transition. set to 0.0f for cross fade. set to -1.0f for fade out and fade in */
	private float mAlpha = 1.0f;
	// render modes, int8
	private static final int MODE_SPLASH = 0;
	private static final int MODE_NAME = 1;
	private static final int MODE_DIVISIONS = 2;
	private static final int MODE_MENU = 3;
	private static final int MODE_HALL_OF_FAME = 4;
	private static final int MODE_RACE_OPPONENTS = 5;
	private static final int MODE_TRACK_PREVIEW = 6;
	private static final int MODE_RACE = 7;
	private static final int MODE_WAIT_KEY = 8;
	private static final int MODE_SHOW_WON_OR_LOST = 9;
	private static final int MODE_RACE_RESULT = 10;
	private static final int MODE_RACE_RESULT_POINTS = 11;
	private static final int MODE_RACE_RESULT_STATE = 12;
	private static final int MODE_PROMOTION = 13;
	private static final int MODE_PROMOTION_PICTURE = 14;
	/** current render mode */
	private int mMode = MODE_SPLASH;
	// key state
	static final int KEY_LEFT = 1;
	static final int KEY_RIGHT = 2;
	static final int KEY_UP = 4;
	static final int KEY_DOWN = 8;
	static final int KEY_FIRE = 16;
	static final int KEY_ROTATE_X_UP = 32;
	static final int KEY_ROTATE_X_DOWN = 64;
	static final int KEY_ROTATE_Y_UP = 128;
	static final int KEY_ROTATE_Y_DOWN = 256;
	private volatile int mKey = 0;
	private volatile int mPrevKey = 0;
	private volatile char mKeyTyped = 0;
	private int mTextPosition = 0;// int8
	// cars
	private final PlayerCar mCar = new PlayerCar();
	private final OpponentCar mOpponentCar = new OpponentCar();
	/** alias for current car view angles, matrix, trigonometric with different coordinates */
	private final Camera mCamera = mCar;
	//
	/** Random number generator */
	static final Random mRandom = new Random( System.currentTimeMillis() );
	/** incremental counter to flash info line */
	private int mFrameCounter = 0;
	/** off road timer */
	private int mOffRoadTimer = 0;// int16
	/** true - practice, false - racing season */
	private boolean mIsPractice = false;
	/** sets after dropping car */
	private boolean mIsRace = false;
	/** true after any car cross the start line */
	private boolean mIsRaceFinished = false;
	/** true - player is winner, false - player is looser */
	private boolean mIsRaceWon = false;
	/** Maximum laps for a race */
	static final int MAX_LAPS = 3;
	/** 3 drivers per division */
	private static final int DRIVERS_IN_DIVISION = 3;
	/** 4 divisions */
	private static final int DIVISION_COUNT = 4;
	/** 2 tracks per division */
	private static final int TRACKS_IN_DIVISION = 2;
	/** current track */
	private Track mTrack = null;
	/** Decoration */
	private Model[] mScenery = null;
	// these variables can be included in one variable, mCurrRaceState = div<<16 | opp<<8 | track
	/** Player division */
	private int mCurrentDivision = DIVISION_COUNT - 1;// 4 division, 0...3, int8
	/** Current opponent for the player in current division, 0 or 1 */
	private int mCurrentOpponent = 0;// 0...1, int8
	/** Current track in current division, 0 or 1 */
	private int mCurrentTrack = 0;// 0...1, int8
	/** Player's data: name and picture */
	private final Driver mPlayer = new Driver("            ".toCharArray(), "p11", 0);
	// text data
	/** Base font size */
	private static final int FONT_SIZE = 9 * PF;
	/** Base font */
	private final Font mFont;
	/** Base font symbol width */
	private int FONT_SYMBOL_WIDTH;
	/** Distance from top to base line for base font */
	private int FONT_BASELINE;
	/** Small font size */
	private static final int SMALLFONT_SIZE = 8 * PF;
	/** Small font */
	private final Font mSmallFont;
	/** Small font symbol width */
	private int SMALLFONT_SYMBOL_WIDTH;
	/** Distance from top to base line for small font */
	private int SMALLFONT_BASELINE;
	/** Available languages */
	private static final String[] LANGUAGES = {
		"en", "ru"
	};
	/** Arabic to Roman from 1 to 4 */
	private static final String[] ROMAN_NUMBERS = { " I", " II", " III", " IV" };
	// Strings to translate
	private static char[] strControl = "Control:".toCharArray();
	private static char[] strLeft = "Left: Steer to left".toCharArray();
	private static char[] strRight = "Right: Steer to right".toCharArray();
	private static char[] strUp = "Up: Accelerate".toCharArray();
	private static char[] strDown = "Down: Brake".toCharArray();
	private static char[] strBoost = "Space, Ctrl, Shift: Boost".toCharArray();
	private static char[] strPause = "Pause: Pause/Resume".toCharArray();
	private static char[] strName = "Name?".toCharArray();
	private static char[] strHallOfFame = "Hall of Fame".toCharArray();
	private static char[] strTRACK = "TRACK".toCharArray();
	private static char[] strDriver = "DRIVER".toCharArray();
	private static char[] strLapTime = "LAP-TIME".toCharArray();
	private static char[] strRaceTime = "RACE-TIME".toCharArray();
	private static char[] strDivision = "Division".toCharArray();
	private static char[] strSelect = "SELECT".toCharArray();
	private static char[] strStartTheRacingSeason = "Start the Racing Season".toCharArray();
	private static char[] strPractise = "Practise".toCharArray();
	private static char[] strLoad = "Load".toCharArray();
	private static char[] strSave = "Save".toCharArray();
	private static char[] strReplay = "Replay".toCharArray();
	private static char[] strLoadSaveReplay = (new String(strLoad) + "/" + new String(strSave) + "/" + new String(strReplay)).toCharArray();
	private static char[] strTracksInDivision1 = "Tracks in DIVISION 1".toCharArray();
	private static char[] strTracksInDivision2 = "Tracks in DIVISION 2".toCharArray();
	private static char[] strTracksInDivision3 = "Tracks in DIVISION 3".toCharArray();
	private static char[] strTracksInDivision4 = "Tracks in DIVISION 4".toCharArray();
	private static char[] strLittleRamp = "LITTLE RAMP".toCharArray();
	private static char[] strHumpBack = "HUMP BACK".toCharArray();
	private static char[] strSteppingStones = "STEPPING STONES".toCharArray();
	private static char[] strBigRamp = "BIG RAMP".toCharArray();
	private static char[] strHighJump = "HIGH JUMP".toCharArray();
	private static char[] strRollerCoaster = "ROLLER COASTER".toCharArray();
	private static char[] strSkiJump = "SKI JUMP".toCharArray();
	private static char[] strDrawbridge = "DRAWBRIDGE".toCharArray();
	private static char[] strPractiseTheLittleRamp = "Practise The LITTLE RAMP".toCharArray();
	private static char[] strPractiseTheHumpBack = "Practise The HUMP BACK".toCharArray();
	private static char[] strPractiseTheSteppingStones = "Practise The STEPPING STONES".toCharArray();
	private static char[] strPractiseTheBigRamp = "Practise The BIG RAMP".toCharArray();
	private static char[] strPractiseTheHighJump = "Practise The HIGH JUMP".toCharArray();
	private static char[] strPractiseTheRollerCoaster = "Practise The ROLLER COASTER".toCharArray();
	private static char[] strPractiseTheSkiJump = "Practise The SKI JUMP".toCharArray();
	private static char[] strPractiseTheDrawbridge = "Practise The DRAWBRIDGE".toCharArray();
	private static char[] strCancel = "Cancel".toCharArray();
	private static char[] strDrop = "DROP START".toCharArray();
	private static char[] strRace = "RACE".toCharArray();
	private static char[] strOf = "of".toCharArray();
	private static char[] strTrack = "Track:".toCharArray();
	private static char[] strSteerToRotateView = "steer to rotate view or fire to continue".toCharArray();
	private static char[] strNewTrackRecords = "New track records".toCharArray();
	private static char[] strResult = "RESULT".toCharArray();
	private static char[] strFirst = "First".toCharArray();
	private static char[] strSecond = "Second".toCharArray();
	private static char[] strThird = "Third".toCharArray();
	private static char[] strRaced = "Raced".toCharArray();
	private static char[] strWins = "Wins".toCharArray();
	private static char[] strWreck = "WRECK".toCharArray();
	private static char[] strLaps = "Laps".toCharArray();
	private static char[] strLapsOver = "LAPS OVER".toCharArray();
	private static char[] strRaceWon = "RACE WON".toCharArray();
	private static char[] strRaceLost = "RACE LOST".toCharArray();
	private static char[] strPressFire = "PRESS FIRE".toCharArray();
	private static char[] strPoints = "Points".toCharArray();
	private static char[] strChanges = "CHANGES".toCharArray();
	private static char[] strPromotion = "Promotion for".toCharArray();
	private static char[] strRelegation = "Relegation for".toCharArray();
	private static char[] strResultRaceTime = "Race Time:".toCharArray();
	private static char[] strResultBestLap = "Best Lap :".toCharArray();
	private static char[] strWinnerPts = "Winner 2pts".toCharArray();
	private static char[] strBestLapPts = "Best Lap 1pt".toCharArray();
	// Racer's names, 12 chars maximum
	private static char[] strHotRod = "Hot Rod".toCharArray();
	private static char[] strWhizzKid = "Whizz Kid".toCharArray();
	private static char[] strBadGuy = "Bad Guy".toCharArray();
	private static char[] strDodger = "The Dodger".toCharArray();
	private static char[] strBigEd = "Big Ed".toCharArray();
	private static char[] strMaxBoost = "Max Boost".toCharArray();
	private static char[] strDareDevil = "Dare Devil".toCharArray();
	private static char[] strHighFlyer = "High Flyer".toCharArray();
	private static char[] strBullyBoy = "Bully Boy".toCharArray();
	private static char[] strJumpinJack = "Jumpin' Jack".toCharArray();
	private static char[] strRoadHog = "Road Hog".toCharArray();
	/** Division data: tracks, drivers */
	private final Division[] mDivisions = {
		new Division(// division 1
			new Track[] { new TrackDrawBridge(), new TrackSkiJump() },
			new Driver[] {
				new Driver(strHotRod, "p0", OpponentCar.STYLE_PUSH_PLAYER | OpponentCar.STYLE_OBSTRUCTS_PLAYER),
				new Driver(strWhizzKid, "p1", OpponentCar.STYLE_PUSH_PLAYER),
				new Driver(strBadGuy, "p2", OpponentCar.STYLE_PUSH_PLAYER | OpponentCar.STYLE_OBSTRUCTS_PLAYER)
			}
		),
		new Division(// division 2
			new Track[] { new TrackRollerCoaster(), new TrackHighJump() },
			new Driver[] {
				new Driver(strDodger, "p3", OpponentCar.STYLE_PUSH_PLAYER),
				new Driver(strBigEd, "p4", OpponentCar.STYLE_PUSH_PLAYER | OpponentCar.STYLE_OBSTRUCTS_PLAYER | OpponentCar.STYLE_DRIVES_NEAR_EDGE | OpponentCar.STYLE_WHEELIE),
				new Driver(strMaxBoost, "p5", OpponentCar.STYLE_WHEELIE)
			}
		),
		new Division(// division 3
			new Track[] { new TrackBigRamp(), new TrackSteppingStones() },
			new Driver[] {
				new Driver(strDareDevil, "p6", OpponentCar.STYLE_PUSH_PLAYER),
				new Driver(strHighFlyer, "p7", OpponentCar.STYLE_WHEELIE),
				new Driver(strBullyBoy, "p8", OpponentCar.STYLE_DRIVES_NEAR_EDGE | OpponentCar.STYLE_OBSTRUCTS_PLAYER)
			}
		),
		new Division(// division 4
			new Track[] { new TrackHumpBack(), new TrackLittleRamp() },
			new Driver[] {
				new Driver(strJumpinJack, "p9", 0),
				new Driver(strRoadHog, "p10", OpponentCar.STYLE_DRIVES_NEAR_EDGE),
				mPlayer
			}
		)
	};
	/** Space to store current division states to be able do replay */
	private final Driver[][] mBackupDivisionState = new Driver[DIVISION_COUNT][DRIVERS_IN_DIVISION];
/*
	final char[][] strShortTrackNames = {
		strDrawbridge, strSkiJump,
		strRollerCoaster, strHighJump,
		strBigRamp, strSteppingStones,
		strHumpBack, strLittleRamp
	};
*/
	private static final char[][] menuMain = {
		strHallOfFame, strPractise, strStartTheRacingSeason, strLoadSaveReplay
	};
	private static final char[][] menuPractise = {
		strTracksInDivision1, strTracksInDivision2, strTracksInDivision3, strTracksInDivision4
	};
	private static final char[][] menuTrackSelectDiv1 = {
		strPractiseTheSkiJump, strPractiseTheDrawbridge, strCancel
	};
	private static final char[][] menuTrackSelectDiv2 = {
		strPractiseTheHighJump, strPractiseTheRollerCoaster, strCancel
	};
	private static final char[][] menuTrackSelectDiv3 = {
		strPractiseTheSteppingStones, strPractiseTheBigRamp, strCancel
	};
	private static final char[][] menuTrackSelectDiv4 = {
		strPractiseTheLittleRamp, strPractiseTheHumpBack, strCancel
	};
	private static final char[][] menuLoadSaveReplay = {
		strLoad, strSave, strReplay, strCancel
	};
	private char[][] mMenu = menuMain;
	private int mMenuSelectedItem = 0;// int8
	//
	static void main(String args[]) {
		SwingUtilities.invokeLater( new Runnable() {
			public void run() {
				Thread.currentThread().setName( "Game" );
				new GameView().init();
			}
		});
	}
	GameView() {
		setName( APP_NAME );
		setTitle( APP_NAME );
		setDefaultCloseOperation( JFrame.EXIT_ON_CLOSE );
		addWindowListener( new WindowListener() {
			public void windowActivated(WindowEvent e) {
					//System.out.println("Window Activated");
				}
				public void windowOpened(WindowEvent e) {
					//System.out.println("Window Opened");
				}
				public void windowClosing(WindowEvent e) {
					//System.out.println("Window Closing");
					stop();
					setVisible( false );
					dispose();
					if( mImgOffscreenBuffer != null ) {
						mImgOffscreenBuffer.flush();
						mG.dispose();
					}
					System.exit( 0 );
				}
				public void windowClosed(WindowEvent e) {
					//System.out.println("Window close event occur");
				}
				public void windowDeactivated(WindowEvent e) {
					//System.out.println("Window Deactivated");
				}
				public void windowDeiconified(WindowEvent e) {
					//System.out.println("Window Deiconified");
				}
				public void windowIconified(WindowEvent e) {
					//System.out.println("Window Iconified");
				}
			}
		);
		pack();
		int width = SCREEN_WIDTH;
		int height = SCREEN_HEIGHT;
		final Rectangle r = getGraphicsConfiguration().getBounds();
		while( width > r.width || height > r.height ) {
			width >>= 1;
			height >>= 1;
		}
		final Insets in = getInsets();
		setSize( width + in.left + in.right, height + in.top + in.bottom );
		setLocationRelativeTo( null );// center window on screen
		setResizable( true );//false );
		setFocusable( true );
		addKeyListener( this );
		addComponentListener( this );
		setIgnoreRepaint( true );
		setVisible( true );
		// initiate replay data
		saveState();
		//
		Font font;
		Font small_font;
		InputStream is = null;
		try {
			is = GameView.class.getResourceAsStream("Lucida Console Regular.ttf");
			Font f = Font.createFont( Font.TRUETYPE_FONT, is );
			font = f.deriveFont( Font.PLAIN, FONT_SIZE );
			small_font = f.deriveFont( Font.PLAIN, SMALLFONT_SIZE );
		} catch( Exception e ) {
			font = new Font( Font.MONOSPACED, Font.PLAIN, FONT_SIZE );
			small_font = new Font( Font.MONOSPACED, Font.PLAIN, SMALLFONT_SIZE );
			//System.out.println("Font don't found, will use default");
			System.out.println( e.getMessage() );
		} finally {
			if( is != null ) try { is.close(); } catch (IOException e) {}
		}
		mFont = font;
		mSmallFont = small_font;
	}

	private void createOffscreenBuffer() {
		if( mImgOffscreenBuffer != null ) {
			mImgOffscreenBuffer.flush();
			mG.dispose();
		}
		//getGraphicsConfiguration().createCompatibleVolatileImage( width, height );
		mImgOffscreenBuffer = createVolatileImage( SCREEN_WIDTH, SCREEN_HEIGHT );
		//mOffscreenBuffer = createImage( width, height );
		if( mImgOffscreenBuffer != null ) {
			mG = mImgOffscreenBuffer.getGraphics();
			FontMetrics fm = mG.getFontMetrics( mFont );
			FONT_SYMBOL_WIDTH = fm.charWidth(' ');
			// baseline = top+((bottom+1-top)/2)-((ascent+descent)/2)+ascent;
			// to center:
			// baseline = (bottom + 1 + top) / 2 + (ascent - descent) / 2
			// to print with a left top anchor point:
			// baseline = (top+size + 1 + top) / 2 + (ascent - descent) / 2
			// baseline = top + (size + 1 + ascent - descent) / 2
			FONT_BASELINE = (FONT_SIZE + 1 + fm.getAscent() - fm.getDescent()) >> 1;
			//
			fm = mG.getFontMetrics( mSmallFont );
			SMALLFONT_SYMBOL_WIDTH = fm.charWidth(' ');
			SMALLFONT_BASELINE = (SMALLFONT_SIZE + 1 + fm.getAscent() - fm.getDescent()) >> 1;
		}
	}

	/* don not work with non ascii symbols
	private char[] getString(final ResourceBundle rb, final String key, final char[] default_string) {
		try {
			return rb.getString( key ).toCharArray();
		} catch (MissingResourceException mre ) {
			System.out.println( mre.getMessage() );
		}
		return default_string;
	}
	*/
	private char[] getString(final Properties rb, final String key, final char[] default_string) {
		String str = rb.getProperty( key );
		if( str == null ) {
			System.out.println("Don't found string for: " + key);
			return default_string;
		}
		return str.toCharArray();
	}
	/**
	 * Internal dialog localization
	 * @param name resource name
	 * @param language name of a language
	 */
	private static final void loadUIManagerLocalization(final String name, final String language) {
		InputStreamReader ir = null;
		try {
			final InputStream is = GameView.class.getResourceAsStream( name + "_" + language + ".txt" );
			if( is == null ) {
				return;
			}
			ir = new InputStreamReader( is, TEXT_ENCODE );
			final Properties p = new Properties();
			p.load( ir );
			for( final Object k : p.keySet() ) {
				UIManager.put( (String)k, p.getProperty( (String)k ) );
			}
		} catch(final UnsupportedEncodingException e) {
		} catch(final IOException e) {
		} finally {
			if( ir != null ) {
				try { ir.close(); } catch( final IOException e ) {}
			}
		}
	}
	void setLanguage(final String language) {
synchronized( mPaintLock ) {// practically do not needed
		loadUIManagerLocalization( "FileChooser", language );
		InputStreamReader ir = null;
		try {
			//ResourceBundle rb = ResourceBundle.getBundle( language );// do not work for non ascii symbols
			ir = new InputStreamReader( GameView.class.getResourceAsStream( language + ".properties" ), TEXT_ENCODE );
			Properties res = new Properties();
			res.load( ir );
			//
			strControl = getString( res, "strControl", strControl );
			strLeft = getString( res, "strLeft", strLeft );
			strRight = getString( res, "strRight", strRight );
			strUp = getString( res, "strUp", strUp );
			strDown = getString( res, "strDown", strDown );
			strBoost = getString( res, "strBoost", strBoost );
			strPause = getString( res, "strPause", strPause );
			strName = getString( res, "strName", strName );
			strHallOfFame = getString( res, "strHallOfFame", strHallOfFame );
			strTRACK = getString( res, "strTRACK", strTRACK );
			strDriver = getString( res, "strDriver", strDriver );
			strLapTime = getString( res, "strLapTime", strLapTime );
			strRaceTime = getString( res, "strRaceTime", strRaceTime );
			strDivision = getString( res , "strDivision", strDivision );
			strSelect = getString( res , "strSelect", strSelect );
			strStartTheRacingSeason = getString( res, "strStartTheRacingSeason", strStartTheRacingSeason );
			strPractise = getString( res, "strPractise", strPractise );
			strLoad = getString( res, "strLoad", strLoad );
			strSave = getString( res, "strSave", strSave );
			strReplay = getString( res, "strReplay", strReplay );
			strLoadSaveReplay = (new String(strLoad) + "/" + new String(strSave) + "/" + new String(strReplay)).toCharArray();
			strTracksInDivision1 = getString( res, "strTracksInDivision1", strTracksInDivision1 );
			strTracksInDivision2 = getString( res, "strTracksInDivision2", strTracksInDivision2 );
			strTracksInDivision3 = getString( res, "strTracksInDivision3", strTracksInDivision3 );
			strTracksInDivision4 = getString( res, "strTracksInDivision4", strTracksInDivision4 );
			strLittleRamp = getString( res, "strLittleRamp", strLittleRamp );
			strHumpBack = getString( res, "strHumpBack", strHumpBack );
			strSteppingStones = getString( res, "strSteppingStones", strSteppingStones );
			strBigRamp = getString( res, "strBigRamp", strBigRamp );
			strHighJump = getString( res, "strHighJump", strHighJump );
			strRollerCoaster = getString( res, "strRollerCoaster", strRollerCoaster );
			strSkiJump = getString( res, "strSkiJump", strSkiJump );
			strDrawbridge = getString( res, "strDrawbridge", strDrawbridge );
			strPractiseTheLittleRamp = getString( res, "strPractiseTheLittleRamp", strPractiseTheLittleRamp );
			strPractiseTheHumpBack = getString( res, "strPractiseTheHumpBack", strPractiseTheHumpBack );
			strPractiseTheSteppingStones = getString( res, "strPractiseTheSteppingStones", strPractiseTheSteppingStones );
			strPractiseTheBigRamp = getString( res, "strPractiseTheBigRamp", strPractiseTheBigRamp );
			strPractiseTheHighJump = getString( res, "strPractiseTheHighJump", strPractiseTheHighJump );
			strPractiseTheRollerCoaster = getString( res, "strPractiseTheRollerCoaster", strPractiseTheRollerCoaster );
			strPractiseTheSkiJump = getString( res, "strPractiseTheSkiJump", strPractiseTheSkiJump );
			strPractiseTheDrawbridge = getString( res, "strPractiseTheDrawbridge", strPractiseTheDrawbridge );
			strCancel = getString( res, "strCancel", strCancel );
			strDrop = getString( res, "strDrop", strDrop );
			strRace = getString( res, "strRace", strRace );
			strOf = getString( res, "strOf", strOf );
			strTrack = getString( res, "strTrack", strTrack );
			strSteerToRotateView = getString( res, "strSteerToRotateView", strSteerToRotateView );
			strNewTrackRecords = getString( res, "strNewTrackRecords", strNewTrackRecords );
			strResult = getString( res, "strResult", strResult );
			strFirst = getString( res, "strFirst", strFirst );
			strSecond = getString( res, "strSecond", strSecond );
			strThird = getString( res, "strThird", strThird );
			strRaced = getString( res, "strRaced", strRaced );
			strWins = getString( res, "strWins", strWins );
			strWreck = getString( res, "strWreck", strWreck );
			strLaps = getString( res, "strLaps", strLaps );
			strLapsOver = getString( res, "strLapsOver", strLapsOver );
			strRaceLost = getString( res, "strRaceLost", strRaceLost );
			strRaceWon = getString( res, "strRaceWon", strRaceWon );
			strPressFire = getString( res, "strPressFire", strPressFire );
			strPoints = getString( res, "strPoints", strPoints );
			strChanges = getString( res, "strChanges", strChanges );
			strPromotion = getString( res, "strPromotion", strPromotion );
			strRelegation = getString( res, "strRelegation", strRelegation );
			strResultRaceTime = getString( res, "strResultRaceTime", strResultRaceTime );
			strResultBestLap = getString( res, "strResultBestLap", strResultBestLap );
			strWinnerPts = getString( res, "strWinnerPts", strWinnerPts );
			strBestLapPts = getString( res, "strBestLapPts", strBestLapPts );
			//
			mDivisions[0].mDrivers[0].setName( getString( res, "strHotRod", strHotRod ) );
			mDivisions[0].mDrivers[1].setName( getString( res, "strWhizzKid", strWhizzKid ) );
			mDivisions[0].mDrivers[2].setName( getString( res, "strBadGuy", strBadGuy ) );
			mDivisions[1].mDrivers[0].setName( getString( res, "strDodger", strDodger ) );
			mDivisions[1].mDrivers[1].setName( getString( res, "strBigEd", strBigEd ) );
			mDivisions[1].mDrivers[2].setName( getString( res, "strMaxBoost", strMaxBoost ) );
			mDivisions[2].mDrivers[0].setName( getString( res, "strDareDevil", strDareDevil ) );
			mDivisions[2].mDrivers[1].setName( getString( res, "strHighFlyer", strHighFlyer ) );
			mDivisions[2].mDrivers[2].setName( getString( res, "strBullyBoy", strBullyBoy ) );
			mDivisions[3].mDrivers[0].setName( getString( res, "strJumpinJack", strJumpinJack ) );
			mDivisions[3].mDrivers[1].setName( getString( res, "strRoadHog", strRoadHog ) );
			/*
			strShortTrackNames[0] = strDrawbridge;
			strShortTrackNames[1] = strSkiJump;
			strShortTrackNames[2] = strRollerCoaster;
			strShortTrackNames[3] = strHighJump;
			strShortTrackNames[4] = strBigRamp;
			strShortTrackNames[5] = strSteppingStones;
			strShortTrackNames[6] = strHumpBack;
			strShortTrackNames[7] = strLittleRamp;
			*/
			menuMain[0] = strHallOfFame;
			menuMain[1] = strPractise;
			menuMain[2] = strStartTheRacingSeason;
			menuMain[3] = strLoadSaveReplay;
			//
			menuPractise[0] = strTracksInDivision1;
			menuPractise[1] = strTracksInDivision2;
			menuPractise[2] = strTracksInDivision3;
			menuPractise[3] = strTracksInDivision4;
			//
			menuTrackSelectDiv1[0] = strPractiseTheSkiJump;
			menuTrackSelectDiv1[1] = strPractiseTheDrawbridge;
			menuTrackSelectDiv1[2] = strCancel;
			//
			menuTrackSelectDiv2[0] = strPractiseTheHighJump;
			menuTrackSelectDiv2[1] = strPractiseTheRollerCoaster;
			menuTrackSelectDiv2[2] = strCancel;
			//
			menuTrackSelectDiv3[0] = strPractiseTheSteppingStones;
			menuTrackSelectDiv3[1] = strPractiseTheBigRamp;
			menuTrackSelectDiv3[2] = strCancel;
			//
			menuTrackSelectDiv4[0] = strPractiseTheLittleRamp;
			menuTrackSelectDiv4[1] = strPractiseTheHumpBack;
			menuTrackSelectDiv4[2] = strCancel;
			//
			menuLoadSaveReplay[0] = strLoad;
			menuLoadSaveReplay[1] = strSave;
			menuLoadSaveReplay[2] = strReplay;
			menuLoadSaveReplay[3] = strCancel;
		} catch( Exception e ) {
			System.out.println( e.getMessage() );
		} finally {
			if( ir != null ) try { ir.close(); } catch( IOException e ) {}
		}
}
	}
	/** Save current division states to be able to do replay */
	private void saveState() {
		for( int d = 0; d < DIVISION_COUNT; d++ ) {
			System.arraycopy( mDivisions[d].mDrivers, 0, mBackupDivisionState[d], 0, DRIVERS_IN_DIVISION );
		}
	}
	/** Save game state: hall of fame and divisions */
	private void save(final File file) {
		// save to file
		BufferedWriter bwr = null;
		try {
			bwr = new BufferedWriter( new OutputStreamWriter( new FileOutputStream( file ), TEXT_ENCODE ) );
			bwr.write( "# " + APP_NAME );
			bwr.newLine();
			for( int d = 0; d < DIVISION_COUNT; d++ ) {
				final Division division = mDivisions[d];
				// hall of fame
				for( int t = 0; t < TRACKS_IN_DIVISION; t++ ) {
					final Track track = division.mTracks[t];
					final String prefix = "div" + d + ".track" + t + ".";
					bwr.write( prefix + "lapname = " + new String( track.mBestLapPilot ) );
					bwr.newLine();
					bwr.write( prefix + "laptime = " + Integer.toString( track.mBestLapTimeMs10 ) );
					bwr.newLine();
					bwr.write( prefix + "racename = " + new String( track.mBestRacePilot ) );
					bwr.newLine();
					bwr.write( prefix + "racetime = " + Integer.toString( track.mBestRaceTimeMs10 ) );
					bwr.newLine();
				}
				// divisions
				for( int n = 0; n < DRIVERS_IN_DIVISION; n++ ) {
					bwr.write( "div" + d + ".driver" + n + " = " + division.mDrivers[n].mPictureName );
					bwr.newLine();
				}
			}
		} catch( IOException e ) {
			System.out.println( e.getMessage() );
			e.printStackTrace();
		} finally {
			if( bwr != null ) try { bwr.close(); } catch( IOException e ) {}
		}
	}
	/** Load game state: hall of fame and divisions */
	private int load(final File file) {
		if( file.length() > 2048 ) return -1;
		BufferedReader brd = null;
		try {
			brd = new BufferedReader( new InputStreamReader( new FileInputStream( file ), TEXT_ENCODE ) );
			Properties p = new Properties();
			p.load( brd );
			brd.close();
			brd = null;
			// test values
			for( int d = 0; d < DIVISION_COUNT; d++ ) {
				// hall of fame
				for( int t = 0; t < TRACKS_IN_DIVISION; t++ ) {
					final String prefix = "div" + d + ".track" + t + ".";
					String key = prefix + "lapname";
					String val = p.getProperty( key );
					if( val == null || val.length() > 12 ) return -1;
					key = prefix + "laptime";
					int time = Integer.parseInt( p.getProperty( key, "" ) );
					if( time != 0x00ffffff && (time < 10*1000/10 || time > 9*60*1000/10) ) return -1;
					key = prefix + "racename";
					val = p.getProperty( key );
					if( val == null || val.length() > 12 ) return -1;
					key = prefix + "racetime";
					time = Integer.parseInt( p.getProperty( key, "" ) );
					if( time != 0x00ffffff && (time < 10*1000/10 || time > 9*60*1000/10) ) return -1;
				}
				// divisions
				for( int n = 0; n < DRIVERS_IN_DIVISION; n++ ) {
					String key = "div" + d + ".driver" + n;
					String val = p.getProperty( key, "" );
					if( val.length() > 3 ) return -1;
					if( val.charAt( 0 ) != 'p' ) return -1;
					int num = Integer.parseInt( val.substring( 1 ) );
					if( num < 0 || num > (DIVISION_COUNT * DRIVERS_IN_DIVISION - 1) ) return -1;
				}
			}
			// load values
			for( int d = 0; d < DIVISION_COUNT; d++ ) {
				final Division division = mDivisions[d];
				// hall of fame
				for( int t = 0; t < TRACKS_IN_DIVISION; t++ ) {
					final Track track = division.mTracks[t];
					final String prefix = "div" + d + ".track" + t + ".";
					String key = prefix + "lapname";
					String val = p.getProperty( key );
					Track.setName( track.mBestLapPilot, val.toCharArray() );
					key = prefix + "laptime";
					val = p.getProperty( key );
					track.mBestLapTimeMs10 = Integer.parseInt( val );
					key = prefix + "racename";
					val = p.getProperty( key );
					Track.setName( track.mBestRacePilot, val.toCharArray() );
					key = prefix + "racetime";
					val = p.getProperty( key );
					track.mBestRaceTimeMs10 = Integer.parseInt( val );
				}
				// divisions
				for( int n = 0; n < DRIVERS_IN_DIVISION; n++ ) {
					String key = "div" + d + ".driver" + n;
					String val = p.getProperty( key );
					for( int ds = 0; ds < DIVISION_COUNT; ds++ ) {
						final Division scan_div = mDivisions[ds];
						for( int ns = 0; ns < DRIVERS_IN_DIVISION; ns++ ) {
							final Driver scan_driver = scan_div.mDrivers[ns];
							if( scan_driver.mPictureName.equals( val ) ) {
								Driver driver = division.mDrivers[n];
								division.mDrivers[n] = scan_driver;
								scan_div.mDrivers[ns] = driver;
								if( val.equals( mPlayer.mPictureName ) ) {
									mCurrentDivision = d;
								}
							}
						}
					}
				}
			}
		} catch( IOException e ) {
			System.out.println( e.getMessage() );
			e.printStackTrace();
		} finally {
			if( brd != null ) try { brd.close(); } catch( IOException e ) {}
		}
		return 0;
	}
	void pause() {
		synchronized( mPauseLock ) {
			mIsPaused = true;
		}
		SoundManager.stopAll();
	}
	void resume() {
		synchronized( mPauseLock ) {
			mIsPaused = false;
			mPauseLock.notifyAll();
		}
		SoundManager.restart();
	}
	void stop() {
		synchronized( mPauseLock ) {
			mIsPaused = false;
			mIsRunning = false;
			mPauseLock.notifyAll();
		}
		SoundManager.release();
	}

	@Override
	public void componentHidden(ComponentEvent e) {
		//System.out.println("componentHidden event occur");
	}
	@Override
	public void componentMoved(ComponentEvent e) {
	}
	@Override
	public void componentResized(ComponentEvent e) {
		//System.out.println("componentResized event occur: w = " + getWidth() + ", h = " + getHeight());
		final Insets inset = getInsets();
		mPaintScaleWidth = ((double)(getWidth() - inset.left - inset.right) / (double)SCREEN_WIDTH);
		mPaintScaleHeight = ((double)(getHeight() - inset.top - inset.bottom) / (double)SCREEN_HEIGHT);
		mFrameViewportTopMargin = (int)(inset.top / mPaintScaleHeight);
		mFrameViewportLeftMargin = (int)(inset.left / mPaintScaleWidth);
	}
	@Override
	public void componentShown(ComponentEvent e) {
		//System.out.println("componentShown event occur");
	}
	/** Load an image with given name from image folder and correct extension
	 * @param name - a image name without extension
	 * @return InputStream to read the image
	 * @throws IOException 
	 */
	static BufferedImage getImage(final String name) throws IOException {
		return ImageIO.read( GameView.class.getResource( "images/" + name + ".png" ) );
	}
	/** load resources */
	void init() {
		if( GraphicsEnvironment.isHeadless() ) return;
		// set icon
		try { setIconImage( getImage("ico") ); } catch(IOException ie) {}
		//
		if( Dashboard.init() < 0 )
			return;
		//
		if( SoundManager.init() < 0 )
			return;
		//
		mIsRunning = true;
		mThread = new Thread( this, "GameLoop" );
		mThread.start();
	}
	/**
	 * Load and render an image with given name at 0, 0
	 * @param g
	 * @param name
	 */
	private void renderBackgroundImage(final Graphics g, final String name) {
		try {
			Image img = getImage( name ).getScaledInstance( SCREEN_WIDTH, SCREEN_HEIGHT, BufferedImage.SCALE_SMOOTH );
			g.drawImage( img, 0, 0, null );
		} catch(Exception e) {
			System.out.println( e.getMessage() );
			e.printStackTrace();
		}
	}
	/**
	 * Load and render a flag
	 * @param g
	 * @param name image resource name without extension
	 * @param x x coordinate
	 * @param y y coordinate
	 * @param isSelected true - selected, false otherwise
	 */
	private void renderLangugeItem(final Graphics g, final String name, int x, int y, boolean isSelected) {
		final int image_width = 24 * PF;
		final int image_height = 19 * PF;
		//
		if( isSelected ) {
			g.setColor( Color.RED );
			g.fillRect( x - 1 * PF, y - 1 * PF, image_width + 2 * PF, image_height + 2 * PF );
		}
		//
		try {
			Image img = getImage( name ).getScaledInstance( image_width, image_height, BufferedImage.SCALE_SMOOTH );
			g.drawImage( img, x, y, null );
		} catch(Exception e) {
			System.out.println( e.getMessage() );
			e.printStackTrace();
		}
	}
	/**
	 * Render splash screen and menu to choose language
	 * @param g
	 */
	private void renderSplash(final Graphics g) {
		// show picture
		renderBackgroundImage( g, "splash" );
		// show controls
		final int bottom_margin = 8 * PF;
		final int hor_margin = 5 * PF;
		final int ver_margin = SCREEN_HEIGHT - bottom_margin - 7 * SMALLFONT_SIZE + SMALLFONT_BASELINE;
		g.setColor( new Color( 128, 160, 128 ) );
		g.setFont( mSmallFont );
		g.drawChars( strControl, 0, strControl.length, hor_margin, ver_margin );
		g.drawChars( strLeft, 0, strLeft.length, hor_margin, ver_margin + SMALLFONT_SIZE );
		g.drawChars( strRight, 0, strRight.length, hor_margin, ver_margin + 2 * SMALLFONT_SIZE );
		g.drawChars( strUp, 0, strUp.length, hor_margin, ver_margin + 3 * SMALLFONT_SIZE );
		g.drawChars( strDown, 0, strDown.length, hor_margin, ver_margin + 4 * SMALLFONT_SIZE );
		g.drawChars( strBoost, 0, strBoost.length, hor_margin, ver_margin + 5 * SMALLFONT_SIZE );
		g.drawChars( strPause, 0, strPause.length, hor_margin, ver_margin + 6 * SMALLFONT_SIZE );
		// show language
		final int image_width = 24;
		final int image_height = 19;
		final int language_y = ver_margin - 2 * SMALLFONT_SIZE - image_height * PF;
		//
		renderLangugeItem( g, LANGUAGES[0], 5 * PF, language_y, mMenuSelectedItem == 0 );
		renderLangugeItem( g, LANGUAGES[1], (5 + 5 + image_width) * PF, language_y, mMenuSelectedItem == 1 );
	}
	/**
	 * Render menu line with decoration
	 * @param g - Graphics
	 * @param color - background color for the line
	 * @param y - vertical coordinate
	 * @param text - text to draw
	 */
	private void renderMenuLine(final Graphics g, final Color color, int y, final char[] text) {
		final Color color1 = new Color( 224, 224, 224 );
		final Color color2 = new Color( 64, 64, 64 );
		final int line_width = 224 * PF;
		final int line_height = FONT_SIZE * 2;
		final int left_top_x = 32 * PF;
		final int nail_size = 2 * PF;
		final int nail_shift = 2 * PF;
		// draw background
		g.setColor( color );
		g.fillRect( left_top_x, y, line_width, line_height );
		g.setColor( color1 );
		g.fillRect( left_top_x + 1 * PF, y, line_width - 2 * PF, 1 * PF );
		g.setColor( Color.BLACK );
		g.fillRect( left_top_x + 1 * PF, y + (line_height - 1 * PF), line_width - 2 * PF, 1 * PF );
		// draw nails
		g.setColor( color2 );
		g.fillRect( left_top_x + nail_shift, y + (line_height - nail_size) / 2, nail_size, nail_size );
		g.fillRect( left_top_x + line_width - nail_shift - nail_size, y + (line_height - nail_size) / 2, nail_size, nail_size );
		// draw text
		g.setColor( Color.BLACK );
		final int tx = left_top_x + (FONT_SYMBOL_WIDTH << 2);
		g.drawChars( text, 0, text.length, tx, y + ((line_height - FONT_SIZE) / 2 + FONT_BASELINE) );
	}
	/**
	 * Render name request menu.
	 * @param	symbol - UNICODE typed symbol
	 * @return	true if was pressed key "enter", false otherwise
	 */
	private boolean renderNameRequest(final Graphics g, char symbol) {
		final int name_length = 12;
		final char[] player_name = mPlayer.mName;
		if( mTextPosition != 0 && symbol == '\n' ) {
			int t = mTextPosition;
			while( player_name[--t] < ' ' ) ;
			if( ++t < player_name.length ) {
				final char[] name = new char[t];
				System.arraycopy( player_name, 0, name, 0, t );
				mPlayer.mName = name;
			}
			return true;
		}
		final int left_top_x = 32 * PF;
		final int line_width = 224 * PF;
		final int line_height = FONT_SIZE * 2;
		final int text_x = 32 * PF + (FONT_SYMBOL_WIDTH << 2);
		final int prompt_length = FONT_SYMBOL_WIDTH * (name_length + 1);// + length_of(">")
		final int bottom_margin = 8 * PF;
		final int text_y = SCREEN_HEIGHT - (4 - 1) * FONT_SIZE * 3 - bottom_margin;// 4 is max menu lines, 1 is position
		final int underline_x = text_x + FONT_SYMBOL_WIDTH;// + length_of(">")
		g.setFont( mFont );
		final int underline_y = text_y + (line_height - FONT_SIZE) / 2 + FONT_BASELINE + FONT_SIZE / 4;

		if( symbol == '\b' ) {// backspace, delete last symbol
			if( mTextPosition > 0 ) {
				player_name[--mTextPosition] = ' ';
			}
		} else if( (mTextPosition < name_length) && (mTextPosition != 0 || symbol != ' ') ) {
			if( symbol >= ' ' )
				player_name[mTextPosition++] = symbol;
		}
		//
		renderBackgroundImage( g, "menu" );
		//
		g.setColor( new Color( 224, 224, 224 ) );
		final int title_y = text_y - FONT_SIZE * 2 + FONT_BASELINE;
		g.drawChars( strName, 0, strName.length, left_top_x + line_width / 2 - ((strName.length * FONT_SYMBOL_WIDTH) >> 1), title_y );
		//
		final String text = ">" + String.valueOf( player_name );
		renderMenuLine( g, new Color( 160, 160, 160 ), text_y, text.toCharArray() );
		//
		g.drawLine( underline_x, underline_y, text_x + prompt_length, underline_y );
		return false;
	}
	/**
	 * Render current menu. The menu can have 4 lines maximum.
	 * @param g - Graphics
	 * @param mMenu - menu options
	 * @param mMenuSelectedItem - selected line index
	 * @param isDrawDivision - true if need drawing division number
	 * @param isConfirmed - true if need change color for selected line
	 */
	private void renderMenu(final Graphics g, boolean isDrawDivision, boolean isConfirmed) {
		final int left_top_x = 32 * PF;
		final int left_top_y = 70 * PF;
		final int space_halfwidth = 224 * PF / 2;
		//
		final Color color1 = new Color( 160, 160, 160 );
		final Color color2 = new Color( 224, 128, 0 );
		final Color color3 = new Color( 128, 64, 32 );
		final Color color4 = new Color( 224, 224, 224 );
		//
		final int line_height = FONT_SIZE * 2;
		final int space = FONT_SIZE / 2;
		//
		renderBackgroundImage( g, "menu" );
		//
		g.setFont( mFont );
		// current division
		if( isDrawDivision ) {
			g.setColor( color2 );
			String text = String.valueOf( strDivision ) + ROMAN_NUMBERS[mCurrentDivision];
			final int str_division_x = left_top_x + space_halfwidth - ((text.length() * FONT_SYMBOL_WIDTH) >> 1);
			g.drawString( text, str_division_x, left_top_y + FONT_BASELINE );
		}
		// header
		g.setColor( color4 );
		final int str_header_x = left_top_x + space_halfwidth - ((strSelect.length * FONT_SYMBOL_WIDTH) >> 1);
		final int str_header_y = 82 * PF + FONT_BASELINE;
		g.drawChars( strSelect, 0, strSelect.length, str_header_x, str_header_y );
		// draw menu line
		int y = SCREEN_HEIGHT - 4 * line_height - (4 - 1) * space - 16 * PF;// 4 is max menu lines, 16 is bottom margin
		for( int i = 0; i < mMenu.length; i++ ) {
			Color color = color1;
			if( i == mMenuSelectedItem ) {
				color = color2;
				if( isConfirmed )
					color = color3;
			}
			renderMenuLine( g, color, y, mMenu[i] );
			y += (line_height + space);
		}
	}
	private void renderBorder(final Graphics g, int x, int y, int width, int height) {
		g.setColor( Color.BLACK );
		g.drawRect( x - 1, y - 1, width + 1, height + 1 );
		g.setColor( new Color( 128, 128, 96 ) );
		g.drawLine( x - 2, y + height + 1, x + width + 1, y + height + 1 );
		g.drawLine( x + width + 1, y - 2, x + width + 1, y + height );
		g.setColor( new Color( 32, 32, 0 ) );
		g.drawLine( x - 2, y + height, x - 2, y - 1 );
		g.drawLine( x - 2, y - 2, x + width, y - 2 );
	}
	private void renderHallOfFame(final Graphics g) {
		final Color color1 = new Color( 64, 64, 32 );
		final Color color2 = new Color( 160, 160, 160 );
		final Color color3 = new Color( 0, 96, 128 );
		final Color color4 = new Color( 224, 224, 224 );
		final Color color5 = new Color( 64, 0, 0 );
		final Color color6 = new Color( 224, 224, 0 );
		final Color color7 = new Color( 96, 96, 96 );
		// background
		g.setColor( color1 );
		g.fillRect( 0, 0, SCREEN_WIDTH, SCREEN_HEIGHT );
		// show name
		try {
			Image img = getImage("hf").getScaledInstance( 266 * PF, 50 * PF, BufferedImage.SCALE_SMOOTH );
			g.drawImage( img, (SCREEN_WIDTH - 266 * PF) >> 1, 2 * PF, null );
		}
		catch(Exception e) {
			System.out.println( e.getMessage() );
			e.printStackTrace();
		}
		// table header
		final int table_y = 56 * PF;
		final int row_height = FONT_SIZE;
		final int track_name_length = 15;
		final int pilot_name_length = 12;
		final int time_text_length = 7;
		final int row_step_y = (SCREEN_HEIGHT - table_y) / (DIVISION_COUNT * TRACKS_IN_DIVISION + 1);
		final int table_margin_x = 1 * PF;
		//
		renderBorder( g, table_margin_x, table_y, SCREEN_WIDTH - 2 * table_margin_x, row_height );
		g.setColor( color2 );
		g.fillRect( table_margin_x, table_y, SCREEN_WIDTH - 2 * table_margin_x, row_height );
		g.setColor( Color.BLACK );
		// <0.5+15+0.5><0.5+12+0.5><0.5+7+0.5><0.5+12+0.5><0.5+7+0.5>
		int column_space = (SCREEN_WIDTH - (table_margin_x * 2 + (track_name_length + 1 + (pilot_name_length + time_text_length + 2) * 2) * FONT_SYMBOL_WIDTH)) >> 1;
		g.setFont( mFont );
		/*int font_base_line = FONT_BASELINE;
		int font_symbol_width = FONT_SYMBOL_WIDTH;
		if( column_space < 0 ) {
			column_space = 0;
			font_base_line = SMALLFONT_BASELINE;
			font_symbol_width = SMALLFONT_SYMBOL_WIDTH;
			g.setFont( mSmallFont );
		}*/
		final int header_y = table_y + ((row_height - FONT_SIZE) >> 1) + FONT_BASELINE;
		final int name_race_x = SCREEN_WIDTH - table_margin_x - ((pilot_name_length + time_text_length + 2) * FONT_SYMBOL_WIDTH);
		final int race_time_x = name_race_x + pilot_name_length * FONT_SYMBOL_WIDTH + FONT_SYMBOL_WIDTH;
		final int name_lap_x = SCREEN_WIDTH - table_margin_x - ((pilot_name_length + time_text_length + 2) * 2 * FONT_SYMBOL_WIDTH) - column_space;
		final int lap_time_x = name_lap_x + pilot_name_length * FONT_SYMBOL_WIDTH + FONT_SYMBOL_WIDTH;
		//
		g.drawChars( strTRACK, 0, strTRACK.length, table_margin_x + (FONT_SYMBOL_WIDTH >> 1), header_y );
		g.drawChars( strDriver, 0, strDriver.length, name_lap_x + (FONT_SYMBOL_WIDTH >> 1), header_y );
		g.drawChars( strLapTime, 0, strLapTime.length, lap_time_x + (FONT_SYMBOL_WIDTH >> 1) + (time_text_length - strLapTime.length) * FONT_SYMBOL_WIDTH, header_y );
		g.drawChars( strDriver, 0, strDriver.length, name_race_x + (FONT_SYMBOL_WIDTH >> 1), header_y );
		g.drawChars( strRaceTime, 0, strRaceTime.length, race_time_x + (FONT_SYMBOL_WIDTH >> 1) + (time_text_length - strRaceTime.length) * FONT_SYMBOL_WIDTH, header_y );
		//
		int y = table_y + row_step_y;
		for( int d = 0; d < DIVISION_COUNT; d++ ) {
			for( int t = 0; t < TRACKS_IN_DIVISION; t++ ) {
				final Track track = mDivisions[d].mTracks[t];
				// track name
				renderBorder( g, table_margin_x, y, (track_name_length + 1) * FONT_SYMBOL_WIDTH, row_height );
				g.setColor( color3 );
				g.fillRect( table_margin_x, y, (track_name_length + 1) * FONT_SYMBOL_WIDTH, row_height );
				g.setColor( color4 );
				final int text_y = y + FONT_BASELINE;
				g.drawChars( track.mName, 0, track.mName.length, table_margin_x + (FONT_SYMBOL_WIDTH >> 1), text_y );
				// pilot's name + lap time
				renderBorder( g, name_lap_x, y, (pilot_name_length + time_text_length + 2) * FONT_SYMBOL_WIDTH, row_height );
				g.setColor( color5 );
				g.fillRect( name_lap_x, y, pilot_name_length * FONT_SYMBOL_WIDTH + FONT_SYMBOL_WIDTH, row_height );
				g.setColor( color6 );
				g.drawChars( track.mBestLapPilot, 0, pilot_name_length, name_lap_x + (FONT_SYMBOL_WIDTH >> 1), text_y );
				// lap time
				g.setColor( color7 );
				g.fillRect( lap_time_x, y, time_text_length * FONT_SYMBOL_WIDTH + FONT_SYMBOL_WIDTH, row_height );
				g.setColor( color4 );
				Dashboard.convertMsToChars( track.mBestLapTimeMs10 );
				g.drawChars( Dashboard.mStrLapTime, 0, time_text_length, lap_time_x + (FONT_SYMBOL_WIDTH >> 1), text_y );
				// pilot's name + race time
				renderBorder( g, name_race_x, y, (pilot_name_length + time_text_length + 2) * FONT_SYMBOL_WIDTH, row_height );
				g.setColor( color5 );
				g.fillRect( name_race_x, y, pilot_name_length * FONT_SYMBOL_WIDTH + FONT_SYMBOL_WIDTH, row_height );
				g.setColor( color6 );
				g.drawChars( track.mBestRacePilot, 0, pilot_name_length, name_race_x + (FONT_SYMBOL_WIDTH >> 1), text_y );
				// race time
				g.setColor( color7 );
				g.fillRect( race_time_x, y, time_text_length * FONT_SYMBOL_WIDTH + FONT_SYMBOL_WIDTH, row_height );
				g.setColor( color4 );
				Dashboard.convertMsToChars( track.mBestRaceTimeMs10 );
				g.drawChars( Dashboard.mStrLapTime, 0, time_text_length, race_time_x + (FONT_SYMBOL_WIDTH >> 1), text_y );
				//
				y += row_step_y;
			}
		}
	}
	private void renderShadowedBox(final Graphics g, int x, int y, int width, int height) {
		g.setColor( Color.BLACK );
		for( int i = 1; i <= 2 * PF; i += PF ) {
			g.fillRect( x + i, y + i, width, height );
			g.fillRect( x + i + 1, y + i + 1, width, height );
		}
		g.setColor( new Color( 224, 224, 224 ) );
		g.drawRect( x, y, width - 1, height - 1 );
	}
	/**
	 * Current font will be changed
	 */
	private void renderPilotImage(final Graphics g, final Driver driver, int x, int y) {
		final int box_border = 1;
		final int image_width = 72 * PF;
		final int image_height = 43 * PF;
		final int column_width = image_width + box_border * 2;
		final int picture_text_box_height = SMALLFONT_SIZE;
		final int row_height = image_height + picture_text_box_height + box_border * 2;
		//
		renderShadowedBox( g, x, y, column_width, row_height );
		// show picture
		try {
			Image img = getImage( driver.mPictureName ).getScaledInstance( image_width, image_height, BufferedImage.SCALE_SMOOTH );
			g.drawImage( img, x + box_border, y + box_border, null );
		} catch(Exception e) {
			System.out.println( e.getMessage() );
			e.printStackTrace();
		}
		// show pilot's name
		g.setFont( mSmallFont );
		final int tx = x + ((column_width - driver.mName.length * SMALLFONT_SYMBOL_WIDTH) >> 1);
		g.drawChars( driver.mName, 0, driver.mName.length, tx, y + image_height + box_border + SMALLFONT_BASELINE );
	}
	private void renderDivisions(final Graphics g) {
		final int box_border = 1;
		final int box_shadow = 1 * PF;
		final int image_width = 72 * PF;
		final int image_height = 43 * PF;
		final int column_width = image_width + box_border * 2;
		final int column_space = 6 * PF;
		final int division_text_box_height = SMALLFONT_SIZE;
		final int picture_text_box_height = SMALLFONT_SIZE;
		final int row_height = image_height + picture_text_box_height + box_border * 2;
		final int row_space = box_shadow + 1 * PF;
		final int tracks_text_box_height = SMALLFONT_SIZE * 2;
		final int left_top_x = SCREEN_HALFWIDTH -
				(((column_width + column_space) * DIVISION_COUNT - column_space + box_shadow) / 2);
		final int left_top_y = SCREEN_HALFHEIGHT -
				((row_height + row_space) * DRIVERS_IN_DIVISION + 
				(division_text_box_height + box_border * 2) + row_space +
				(tracks_text_box_height + box_border * 2) + box_shadow) / 2;
		// colors
		final Color color1 = new Color( 224, 224, 224 );
		final Color color2 = new Color( 64, 64, 64 );
		// font
		g.setFont( mSmallFont );
		// background
		g.setColor( color2 );
		g.fillRect( 0, 0, SCREEN_WIDTH, SCREEN_HEIGHT );
		//
		for( int d = 0; d < DIVISION_COUNT; d++ ) {
			// show division number
			int x = left_top_x + d * (column_width + column_space);
			renderShadowedBox( g, x, left_top_y, column_width, division_text_box_height + box_border * 2 );
			g.setColor( color1 );
			String text = String.valueOf( strDivision ) + ROMAN_NUMBERS[d];
			int tx = x + ((column_width - text.length() * SMALLFONT_SYMBOL_WIDTH) >> 1);
			g.drawString( text, tx, left_top_y + box_border + SMALLFONT_BASELINE );
			//
			final Division division = mDivisions[d];
			// show pilots
			int y = left_top_y + division_text_box_height + box_border * 2 + row_space;
			for( int i = 0; i < DRIVERS_IN_DIVISION; i++ ) {
				renderPilotImage( g, division.mDrivers[i], x, y );
				y += (row_height + row_space);
			}
			// show track names
			renderShadowedBox( g, x, y, column_width, tracks_text_box_height + box_border * 2 );
			g.setColor( color1 );
			y += box_border + SMALLFONT_BASELINE;
			for( int i = 0; i < TRACKS_IN_DIVISION; i++ ) {
				final char[] track_name = division.mTracks[i].mName;
				tx = x + column_width / 2 - ((track_name.length * SMALLFONT_SYMBOL_WIDTH) >> 1);
				g.drawChars( track_name, 0, track_name.length, tx, y );
				y += SMALLFONT_SIZE;
			}
		}
	}
	/**
	 * 2 lines
	 * @param g
	 * @param color - line background color
	 * @param y - vertical position
	 * @param text1 - text on line 1
	 * @param text2 - text on line 2
	 */
	private void renderInfoLine(final Graphics g, final Color color, int y, final char[] text1, final char[] text2) {
		final Color color1 = new Color( 224, 224, 224 );
		final Color color2 = new Color( 64, 64, 64 );
		final int line_width = 224 * PF;
		final int line_height = FONT_SIZE * 2 + FONT_SIZE;
		final int left_top_x = 32 * PF;
		final int nail_size = 2 * PF;
		final int nail_shift = 2 * PF;
		// draw background
		g.setColor( color );
		g.fillRect( left_top_x, y, line_width, line_height );
		g.setColor( color1 );
		g.fillRect( left_top_x + 2 * PF, y, line_width - 4 * PF, 2 * PF );
		g.setColor( Color.BLACK );
		g.fillRect( left_top_x + 2 * PF, y + (line_height - 2 * PF), line_width - 4 * PF, 2 * PF );
		// draw nails
		g.setColor( color2 );
		g.fillRect( left_top_x + nail_shift, y + (line_height / 3 - nail_size / 2), nail_size, nail_size );
		g.fillRect( left_top_x + line_width - nail_shift - nail_size, y + (line_height / 3 - nail_size / 2), nail_size, nail_size );
		g.fillRect( left_top_x + nail_shift, y + (2 * line_height / 3 - nail_size / 2), nail_size, nail_size );
		g.fillRect( left_top_x + line_width - nail_shift - nail_size, y + (2 * line_height / 3 - nail_size / 2), nail_size, nail_size );
		// draw text
		g.setColor( Color.BLACK );
		final int tx1 = left_top_x + line_width / 2 - ((text1.length * FONT_SYMBOL_WIDTH) >> 1);
		final int tx2 = left_top_x + line_width / 2 - ((text2.length * FONT_SYMBOL_WIDTH) >> 1);
		g.drawChars( text1, 0, text1.length, tx1, y + (line_height / 3 - FONT_SIZE / 2 + FONT_BASELINE) );
		g.drawChars( text2, 0, text2.length, tx2, y + (2 * line_height / 3 - FONT_SIZE / 2 + FONT_BASELINE) );
	}
	/**
	 * Before calling this function current track must be initialized
	 * @param g
	 */
	private void renderRaceOpponents(final Graphics g) {
		final int left_top_x = 32 * PF;
		final int left_top_y = 70 * PF;
		final int space_width = 224 * PF;
		//
		final Color color1 = new Color( 224, 128, 0, 255 );
		final Color color2 = new Color( 224, 224, 224, 255 );
		final Color color3 = new Color( 64, 128, 0, 255 );
		//
		renderBackgroundImage( g, "menu" );
		//
		g.setFont( mFont );
		//
		g.setColor( color1 );
		String text = String.valueOf( strDivision ) + ROMAN_NUMBERS[mCurrentDivision];
		final int str_division_x = left_top_x + space_width / 2 - ((text.length() * FONT_SYMBOL_WIDTH) >> 1);
		g.drawString( text, str_division_x, left_top_y + FONT_BASELINE );
		//
		final int bottom_margin = 8 * PF;
		final int info_y = SCREEN_HEIGHT - bottom_margin - FONT_SIZE * 3;
		//
		text = String.valueOf( strTrack ) + " " + String.valueOf( mTrack.mName );
		final int str_track_x = left_top_x + space_width / 2 - ((text.length() * FONT_SYMBOL_WIDTH) >> 1);
		final int str_track_y = info_y - FONT_SIZE - FONT_SIZE / 2 + FONT_BASELINE;
		g.drawString( text, str_track_x, str_track_y );
		//
		final int race_number = 2 + (mCurrentOpponent << 1) + (TRACKS_IN_DIVISION - mCurrentTrack);// 2 is off screen races
		text = String.valueOf( strRace ) + " " + Integer.toString( race_number ) + " " + String.valueOf( strOf ) + " 6";
		final int str_race_x = left_top_x + space_width / 2 - ((text.length() * FONT_SYMBOL_WIDTH) >> 1);
		final int str_race_y = 82 * PF + FONT_BASELINE;
		g.setColor( color2 );
		g.drawString( text, str_race_x, str_race_y );
		//
		text = "V";
		final int str_v_x = left_top_x + space_width / 2 - ((text.length() * FONT_SYMBOL_WIDTH) >> 1);
		final int str_v_y = 120 * PF + FONT_BASELINE;
		g.drawString( text, str_v_x, str_v_y );
		//
		Dashboard.convertMsToChars( mTrack.mBestRaceTimeMs10 );
		String str_race_time = String.valueOf( strResultRaceTime ) + " " + String.valueOf( mTrack.mBestRacePilot ) + " " + String.valueOf( Dashboard.mStrLapTime );
		Dashboard.convertMsToChars( mTrack.mBestLapTimeMs10 );
		String str_best_lap = String.valueOf( strResultBestLap ) + " " + String.valueOf( mTrack.mBestLapPilot ) + " " + String.valueOf( Dashboard.mStrLapTime );
		renderInfoLine( g, color3, info_y, str_race_time.toCharArray(), str_best_lap.toCharArray() );
		//
		final int image_width = 72 * PF;
		//final int image_height = 43 * PF;
		final int image_y = 93 * PF;
		renderPilotImage( g, mDivisions[mCurrentDivision].mDrivers[mCurrentOpponent], left_top_x + (space_width / 2 - image_width) / 2, image_y );
		renderPilotImage( g, mPlayer, left_top_x + (space_width / 2 - image_width + space_width) / 2, image_y );
	}
	private void renderRaceResult(final Graphics g, final Car car) {
		final int left_top_x = 32 * PF;
		final int left_top_y = 70 * PF;
		final int space_width = 224 * PF;
		//
		final Color color1 = new Color( 224, 128, 0 );
		final Color color2 = new Color( 224, 224, 224 );
		//
		renderBackgroundImage( g, "menu" );
		//
		g.setFont( mFont );
		//
		g.setColor( color1 );
		String text = String.valueOf( strDivision ) + ROMAN_NUMBERS[mCurrentDivision];
		final int str_division_x = left_top_x + space_width / 2 - ((text.length() * FONT_SYMBOL_WIDTH) >> 1);
		g.drawString( text, str_division_x, left_top_y + FONT_BASELINE );
		//
		String str = String.valueOf( strTrack ) + " " + String.valueOf( mTrack.mName );
		final int str_track_x = left_top_x + space_width / 2 - ((str.length() * FONT_SYMBOL_WIDTH) >> 1);
		final int str_track_y = 82 * PF + FONT_BASELINE;
		g.drawString( str, str_track_x, str_track_y );
		//
		final int str_rec_x = left_top_x + space_width / 2 - ((strNewTrackRecords.length * FONT_SYMBOL_WIDTH) >> 1);
		final int str_rec_y = 114 * PF + FONT_BASELINE;
		g.setColor( color2 );
		g.drawChars( strNewTrackRecords, 0, strNewTrackRecords.length, str_rec_x, str_rec_y );
		//
		final int info_y = str_rec_y + FONT_SIZE;
		String str1 = "";
		String str2 = "";
		int time10 = car.mRaceTime / 10;
		if( time10 < mTrack.mBestRaceTimeMs10 ) {
			mTrack.mBestRaceTimeMs10 = time10;
			Dashboard.convertMsToChars( time10 );
			Track.setName( mTrack.mBestRacePilot, mPlayer.mName );
			str1 = String.valueOf( strResultRaceTime ) + " " + String.valueOf( mTrack.mBestRacePilot ) + " " + String.valueOf( Dashboard.mStrLapTime );
		}
		time10 = car.mBestLapTime / 10;
		if( time10 < mTrack.mBestLapTimeMs10 ) {
			mTrack.mBestLapTimeMs10 = time10;
			Dashboard.convertMsToChars( time10 );
			Track.setName( mTrack.mBestLapPilot, mPlayer.mName );
			str2 = String.valueOf( strResultBestLap ) + " " + String.valueOf( mTrack.mBestLapPilot ) + " " + String.valueOf( Dashboard.mStrLapTime );
		}
		renderInfoLine( g, color1, info_y, str1.toCharArray(), str2.toCharArray() );
	}
	private void renderRaceResultPoints(final Graphics g) {
		final int left_top_x = 32 * PF;
		final int left_top_y = 70 * PF;
		final int space_width = 224 * PF;
		//
		final Color color1 = new Color( 224, 128, 0 );
		final Color color2 = new Color( 224, 224, 224 );
		final Color color3 = new Color( 160, 160, 160 );
		//
		renderBackgroundImage( g, "menu" );
		//
		g.setFont( mFont );
		//
		g.setColor( color1 );
		String text = String.valueOf( strDivision ) + ROMAN_NUMBERS[mCurrentDivision];
		final int str_division_x = left_top_x + space_width / 2 - ((text.length() * FONT_SYMBOL_WIDTH) >> 1);
		g.drawString( text, str_division_x, left_top_y + FONT_BASELINE );
		//
		final int race_number = 2 + (mCurrentOpponent << 1) + (TRACKS_IN_DIVISION - mCurrentTrack);// 2 is off screen races
		text = String.valueOf( strRace ) + " " + Integer.toString( race_number ) + " " + String.valueOf( strOf ) + " 6";
		final int str_race_x = left_top_x + space_width / 2 - ((text.length() * FONT_SYMBOL_WIDTH) >> 1);
		final int str_race_y = 79 * PF + FONT_BASELINE;
		g.setColor( color2 );
		g.drawString( text, str_race_x, str_race_y );
		//
		final int str_result_x = left_top_x + space_width / 2 - ((strResult.length * FONT_SYMBOL_WIDTH) >> 1);
		final int str_result_y = 115 * PF + FONT_BASELINE;
		g.drawChars( strResult, 0, strResult.length, str_result_x, str_result_y );
		//
		final int image_width = 72 * PF;
		//final int image_height = 43 * PF;
		//
		char[] chText = strWinnerPts;
		final int str_win_x = left_top_x + space_width / 4 - ((chText.length * FONT_SYMBOL_WIDTH) >> 1);
		final int str_points_y = 127 * PF + FONT_BASELINE;
		g.drawChars( chText, 0, chText.length, str_win_x, str_points_y );
		//
		chText = strBestLapPts;
		final int str_lap_x = left_top_x + space_width / 4 + ((space_width - chText.length * FONT_SYMBOL_WIDTH) >> 1);
		g.drawChars( chText, 0, chText.length, str_lap_x, str_points_y );
		//
		final Driver d1 = mDivisions[mCurrentDivision].mDrivers[mCurrentOpponent];// opponent
		final Driver d2 = mPlayer;// the player
		d1.mRaced++;
		d2.mRaced++;
		//
		final int opponents_y = 90 * PF;
		String str = String.valueOf( d1.mName ) + " V " + String.valueOf( d2.mName );
		renderMenuLine( g, color3, opponents_y, str.toCharArray() );
		//
		final int image_y = 139 * PF;
		// render winner driver image
		Driver d = d1;
		if( mIsRaceWon ) {
			d = d2;
		}
		d.mWins++;
		renderPilotImage( g, d, left_top_x + ((space_width / 2 - image_width) >> 1), image_y );
		// render best lap driver image
		d = d1;
		if( mCar.mBestLapTime < mOpponentCar.mBestLapTime ) {
			d = d2;
		}
		d.mBestLaps++;
		renderPilotImage( g, d, left_top_x + ((space_width / 2 + space_width - image_width) >> 1), image_y );
	}
	private void renderData(final Graphics g, final char[] text, int value, int x, int y) {
		final int line_width = 72 * PF;
		final int margin = FONT_SYMBOL_WIDTH;
		final Color color1 = new Color( 224, 224, 224 );
		final Color color2 = new Color( 224, 128, 0 );
		//
		g.setColor( color2 );
		g.drawChars( text, 0, text.length, x + margin, y );
		g.setColor( color1 );
		String str = Integer.toString( value );
		g.drawString( str, x + line_width - str.length() * FONT_SYMBOL_WIDTH - margin, y );
	}

	private void renderRange(final Graphics g, final char[] range, final Driver driver, int x, int y) {
		final int image_width = 72 * PF;
		final int image_height = 43 * PF;
		//
		g.setFont( mFont );
		//
		g.setColor( new Color( 224, 224, 224 ) );
		y += FONT_BASELINE;
		g.drawChars( range, 0, range.length, x + image_width / 2 - ((range.length * FONT_SYMBOL_WIDTH) >> 1), y );
		//
		y += FONT_SIZE / 2;
		renderPilotImage( g, driver, x, y );
		//
		g.setFont( mFont );
		// raced
		y += image_height + FONT_SIZE + FONT_SIZE + FONT_BASELINE;
		renderData( g, strRaced, driver.mRaced, x, y );
		// wins
		y += FONT_SIZE;
		renderData( g, strWins, driver.mWins, x, y );
		// laps
		y += FONT_SIZE;
		renderData( g, strLaps, driver.mBestLaps, x, y );
		// points
		y += FONT_SIZE;
		renderData( g, strPoints, driver.getPoints(), x, y );
	}

	private void renderRaceResultState(final Graphics g) {
		final int left_top_x = 32 * PF;
		final int left_top_y = 70 * PF;
		final int space_width = 224 * PF;
		final int box_border = 1;
		final int box_shadow = 3;
		//
		renderBackgroundImage( g, "menu" );
		//
		g.setFont( mFont );
		//
		g.setColor( new Color( 224, 128, 0 ) );
		String text = String.valueOf( strDivision ) + ROMAN_NUMBERS[mCurrentDivision];
		final int str_division_x = left_top_x + space_width / 2 - ((text.length() * FONT_SYMBOL_WIDTH) >> 1);
		g.drawString( text, str_division_x, left_top_y + FONT_BASELINE );
		// sorting
		final Driver[] drivers = mDivisions[mCurrentDivision].mDrivers;
		Driver d1 = drivers[0];
		Driver d2 = drivers[1];
		Driver d3 = drivers[2];
		if( d1.getPoints() < d2.getPoints() ) {
			Driver d = d1;
			d1 = d2;
			d2 = d;
		}
		if( d1.getPoints() < d3.getPoints() ) {
			Driver d = d1;
			d1 = d3;
			d3 = d;
		}
		if( d2.getPoints() < d3.getPoints() ) {
			Driver d = d2;
			d2 = d3;
			d3 = d;
		}
		//
		final int image_width = 72 * PF;
		//final int image_height = 43 * PF;
		final int range_y = 80 * PF;
		final int image_box_width = image_width + box_border * 2 + box_shadow;
		renderRange( g, strFirst, d1, left_top_x, range_y );
		renderRange( g, strSecond, d2, left_top_x + ((space_width - image_box_width) >> 1), range_y );
		renderRange( g, strThird, d3, left_top_x + space_width - image_box_width, range_y );
	}
	private void sortDivision(final int division_number) {
		final Driver[] drivers = mDivisions[division_number].mDrivers;
		Driver d1 = drivers[0];
		Driver d2 = drivers[1];
		Driver d3 = drivers[2];
		if( d1.getPoints() < d2.getPoints() ) {
			Driver d = d1;
			d1 = d2;
			d2 = d;
		}
		if( d1.getPoints() < d3.getPoints() ) {
			Driver d = d1;
			d1 = d3;
			d3 = d;
		}
		if( d2.getPoints() < d3.getPoints() ) {
			Driver d = d2;
			d2 = d3;
			d3 = d;
		}
		drivers[0] = d1;
		drivers[1] = d2;
		drivers[2] = d3;
	}
	private void renderPromotion(final Graphics g) {
		final int left_top_x = 32 * PF;
		final int left_top_y = 70 * PF;
		final int space_width = 224 * PF;
		//
		final Color color1 = new Color( 224, 128, 0 );
		final Color color2 = new Color( 224, 224, 224 );
		final Color color3 = new Color( 160, 160, 160 );
		//
		renderBackgroundImage( g, "menu" );
		//
		g.setFont( mFont );
		//
		g.setColor( color1 );
		String text = String.valueOf( strDivision ) + ROMAN_NUMBERS[mCurrentDivision];
		final int str_division_x = left_top_x + space_width / 2 - ((text.length() * FONT_SYMBOL_WIDTH) >> 1);
		g.drawString( text, str_division_x, left_top_y + FONT_BASELINE );
		//
		g.setColor( color2 );
		final int str_changes_x = left_top_x + space_width / 2 - ((strChanges.length * FONT_SYMBOL_WIDTH) >> 1);
		g.drawChars( strChanges, 0, strChanges.length, str_changes_x, left_top_y + FONT_SIZE * 2 + FONT_BASELINE );
		//
		final int line_height = FONT_SIZE * 2;
		final int space = FONT_SIZE / 2;
		//
		final Driver[] drivers = mDivisions[mCurrentDivision].mDrivers;
		// draw promotion
		int line_y = SCREEN_HEIGHT - 3 * line_height - (3 - 1) * space - 8 * PF;// 4 is max menu lines, 8 is bottom margin
		if( mCurrentDivision > 0 ) {
			final String str = String.valueOf( strPromotion ) + "  " + String.valueOf( drivers[0].mName );
			renderMenuLine( g, color3, line_y, str.toCharArray() );
		}
		// draw relegation, if this is not the lowest division
		if( mCurrentDivision < DIVISION_COUNT - 1 ) {
			line_y += (line_height + space);
			final String str = String.valueOf( strRelegation ) + "  " + String.valueOf( drivers[DRIVERS_IN_DIVISION - 1].mName );
			renderMenuLine( g, color3, line_y, str.toCharArray() );
		}
	}
	/** the track map */
	//private int mTrackPoints[] = null;
	/**
	 * Prepare track data, scenery and track map.
	 * @param track - current track
	 */
	private void initTrack(final Track track) {
		track.init();
		mTrack = track;
		Render.setTrack( track );
		/* scaled track map
		final int xcenter = 22 * PF;
		final int ycenter = (60 + 22) * PF;
		final int scale = 600 / PF;
		mTrackPoints = new int[track.mNumTrackCoords];
		int track_point_index = 0;
		int prev_x = -32768;
		int prev_y = -32768;
		for( int n = 0; n < track.mPieces.length; n++ ) {
			final Coord3D[] coords = track.mPieces[n].mCoords;
			final int coords_count = coords.length - 2;
			for( int i = 0; i < coords_count; i += 2 ) {
				int x = coords[i].mX / scale;
				int y = coords[i].mY / scale;
				if( (prev_x != x) || (prev_y != y) ) {
					prev_x = x;
					prev_y = y;
					mTrackPoints[track_point_index++] = xcenter + x;
					mTrackPoints[track_point_index++] = ycenter - y;
				}
			}
		}
		if( mTrackPoints.length > track_point_index ) {
			int[] p = new int[track_point_index];
			for( int n = 0; n < track_point_index; n++ ) { p[n] = mTrackPoints[n]; }
			mTrackPoints = p;
		}
		*/
		// prepare environment
		final Model[] scenery = track.mScenery;
		mScenery = new Model[scenery.length];
		// rotate models
		for( int n = 0; n < mScenery.length; n++ ) {
			mScenery[n] = new Model( scenery[n].copyCoords(), scenery[n].mPolygons )
				.shift( 0, 0, -1 )// shift down to prevent sky showing through between scenery and ground
				.scale( 48, 2, 64 )
				.rotateAroundZ( n * Camera.ANGLE_360 / mScenery.length );
		}
		//
		// set initial preview camera parameters
		//final int distance = (track_size_max * SCALE) / width - CAMERA_FOV;
		//
		mCamera.setAngles( 15 * Camera.ANGLE_360 / 360, 0, 0 );
		mCamera.mX = Track.TRACKS_CENTER_X;
		mCamera.mY = Track.TRACKS_CENTER_Y - 52000;
		mCamera.mZ = 17500;
	}
	@Override
	public void keyPressed(KeyEvent e) {
		switch( e.getKeyCode() )
		{
		case KeyEvent.VK_LEFT:
			mKey |= KEY_LEFT;
			break;
		case KeyEvent.VK_RIGHT:
			mKey |= KEY_RIGHT;
			break;
		case KeyEvent.VK_DOWN:
			mKey |= KEY_DOWN;
			break;
		case KeyEvent.VK_UP:
			mKey |= KEY_UP;
			break;
		case KeyEvent.VK_SPACE:
		case KeyEvent.VK_CONTROL:
		case KeyEvent.VK_SHIFT:
		case KeyEvent.VK_ENTER:
			mKey |= KEY_FIRE;
			break;
		case KeyEvent.VK_PAUSE:
			if( mIsPaused ) resume();
			else pause();
			break;
		case KeyEvent.VK_0:
			mKey |= KEY_ROTATE_X_UP;
			break;
		case KeyEvent.VK_9:
			mKey |= KEY_ROTATE_X_DOWN;
			break;
		case KeyEvent.VK_1:
			mKey |= KEY_ROTATE_Y_UP;
			break;
		case KeyEvent.VK_2:
			mKey |= KEY_ROTATE_Y_DOWN;
			break;
		}
	}
	@Override
	public void keyReleased(KeyEvent e) {
		switch( e.getKeyCode() )
		{
		case KeyEvent.VK_LEFT:
			mKey &= ~KEY_LEFT;
			break;
		case KeyEvent.VK_RIGHT:
			mKey &= ~KEY_RIGHT;
			break;
		case KeyEvent.VK_DOWN:
			mKey &= ~KEY_DOWN;
			break;
		case KeyEvent.VK_UP:
			mKey &= ~KEY_UP;
			break;
		case KeyEvent.VK_SPACE:
		case KeyEvent.VK_CONTROL:
		case KeyEvent.VK_SHIFT:
		case KeyEvent.VK_ENTER:
			mKey &= ~KEY_FIRE;
			break;
		case KeyEvent.VK_0:
			mKey &= ~KEY_ROTATE_X_UP;
			break;
		case KeyEvent.VK_9:
			mKey &= ~KEY_ROTATE_X_DOWN;
			break;
		case KeyEvent.VK_1:
			mKey &= ~KEY_ROTATE_Y_UP;
			break;
		case KeyEvent.VK_2:
			mKey &= ~KEY_ROTATE_Y_DOWN;
			break;
		}
	}
	@Override
	public void keyTyped(KeyEvent e) {
		mKeyTyped = e.getKeyChar();
	}
	private void renderTrackPreview(final Graphics g, final char[] trackName) {
		g.setColor( new Color( 0, 64, 64 ) );
		g.fillRect( 0, 0, SCREEN_WIDTH, SCREEN_HEIGHT );
		// draw decoration
		final int backround_width = 316 * PF;
		final int backround_height = 154 * PF;
		try {
			Image img = getImage("preview").getScaledInstance( backround_width, backround_height, BufferedImage.SCALE_SMOOTH );
			g.drawImage( img, SCREEN_HALFWIDTH - backround_width / 2, 0, null );
		} catch(Exception e) {
			System.out.println( e.getMessage() );
			e.printStackTrace();
		}
		// draw track name
		final int track_name_backround_width = 100 * PF;
		final int track_name_backround_height = FONT_SIZE * 2;
		final int text_y = SCREEN_HEIGHT - FONT_SIZE - FONT_SIZE / 2 + FONT_BASELINE;
		final int track_name_center_y = (backround_height + text_y - FONT_BASELINE) >> 1;
		final Color color1 = new Color( 224, 224, 224 );
		final Color color2 = new Color( 96, 96, 96 );
		final Color color3 = new Color( 224, 224, 0 );
		final Color color4 = new Color( 160, 160, 160 );
		g.setFont( mFont );
		// draw background
		g.setColor( Color.BLACK );
		int x = (SCREEN_WIDTH - track_name_backround_width) / 2;
		int y = track_name_center_y - track_name_backround_height / 2;
		g.fillRect( x, y, track_name_backround_width, FONT_SIZE * 2 );
		g.setColor( color1 );
		x -= 2 * PF;
		y -= 2 * PF;
		g.fillRect( x, y, track_name_backround_width + 2 * PF, 2 * PF );
		g.setColor( color4 );
		y += 2 * PF;
		g.fillRect( x, y, 2 * PF, track_name_backround_height );
		g.fillRect( x + track_name_backround_width, y, 2 * PF, track_name_backround_height );
		g.setColor( color2 );
		y += track_name_backround_height;
		g.fillRect( x, y, track_name_backround_width + 2 * PF, 2 * PF );
		// draw text
		x = (SCREEN_WIDTH - trackName.length * FONT_SYMBOL_WIDTH) >> 1;
		g.setColor( color1 );
		g.drawChars( trackName, 0, trackName.length, x, track_name_center_y - FONT_SIZE / 2 + FONT_BASELINE );
		// description
		g.setColor( color3 );
		x = (SCREEN_WIDTH - strSteerToRotateView.length * FONT_SYMBOL_WIDTH) >> 1;
		g.drawChars( strSteerToRotateView, 0, strSteerToRotateView.length, x, text_y );
		//
		mCamera.calculateZXYmatrix();
		// draw track
		mOpponentCar.mIsRendered = true;// switch off rendering
		//
		Render.sIsPreview = true;
		Render.drawTrack( mG, mCamera, mOpponentCar );
		Render.sIsPreview = false;
	}
	/** Render press key request and reset clip area */
	private void renderKeyRequest(final Graphics g) {
		final int y = 20 * PF;
		final int height = FONT_SIZE * 2;
		final int width = (strPressFire.length + 2) * FONT_SYMBOL_WIDTH;
		g.setColor( new Color( 96, 32, 32) );
		g.fillRoundRect( (SCREEN_WIDTH - width) >> 1, y, width, height, 6 * PF, 6 * PF );
		g.setColor( Color.LIGHT_GRAY );
		g.fillRoundRect( ((SCREEN_WIDTH - width) >> 1) + 2 * PF, y + 2 * PF, width - 4 * PF, height - 4 * PF, 5 * PF, 5 * PF );
		g.setColor( Color.BLACK );
		g.drawChars( strPressFire, 0, strPressFire.length, (SCREEN_WIDTH - strPressFire.length * FONT_SYMBOL_WIDTH) >> 1, y + ((height - FONT_SIZE) >> 1) + FONT_BASELINE );
	}
	/**
	 * @param division - number in range 0...3
	 * @param driver_1 and driver_2 - drivers, which race will be simulated
	 */
	private void generateRaceResult(int division, int driver_1, int driver_2) {
		final Driver[] drivers = mDivisions[division].mDrivers; 
		drivers[driver_1].mRaced = 2;
		drivers[driver_2].mRaced = 2;
		final int wins = mRandom.nextInt( 3 );// 0...2
		drivers[driver_1].mWins = wins;
		drivers[driver_2].mWins = 2 - wins;
		final int best_laps = mRandom.nextInt( 3 );// 0...2
		drivers[driver_1].mBestLaps = best_laps;
		drivers[driver_2].mBestLaps = 2 - best_laps;
	}
	/** Paints the game field */
	public void paint(Graphics g) {
synchronized( mPaintLock ) {
		final Graphics2D g2d = (Graphics2D) g;
		g2d.scale( mPaintScaleWidth, mPaintScaleHeight );
		if( mAlpha < 1.0f ) {
			mAlpha += ((1.0f - 0.0f) * (float)REFRESH_PERIOD / (float)CROSS_FADE_MS);
			if( mAlpha >= 1.0f ) mAlpha = 1.0f;
			float alpha = mAlpha;
			if( alpha < 0.0f ) {
				alpha += 1.0f;
				g2d.setComposite( AlphaComposite.getInstance( AlphaComposite.SRC_OVER, alpha ) );
				g2d.setRenderingHint( RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON );
				g2d.setColor( Color.BLACK );
				g2d.fillRect( mFrameViewportLeftMargin, mFrameViewportTopMargin, SCREEN_WIDTH, SCREEN_HEIGHT );
				return;
			}
			g2d.setComposite( AlphaComposite.getInstance( AlphaComposite.SRC_OVER, alpha ) );
			g2d.setRenderingHint( RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON );
		}
		g2d.drawImage( mImgOffscreenBuffer, mFrameViewportLeftMargin, mFrameViewportTopMargin, null );
}
	}
	private void waiting(long time) {
		try {
			repaint();
			time += REFRESH_PERIOD;
			time -= System.currentTimeMillis();
			if( time > 0 )
				Thread.sleep( time );
		} catch(Exception e){}// if c == null || m_sh = null || interrupt
		finally {
		}
		synchronized( mPauseLock ) {
			if( mIsPaused ) {
				//mSound.pauseAll();
				while( mIsPaused ) {
					try { mPauseLock.wait(); } catch (InterruptedException e) {}
				}
			}
		}
	}
	private void initRace() {
		mIsRace = false;
		mIsRaceFinished = false;
		Dashboard.initRace();
		//
		Car.resetInteraction();
		//
		mCar.setOnCrane( mTrack );
		//
		int style = mDivisions[mCurrentDivision].mDrivers[mCurrentOpponent].mDrivingStyle;
		if( mRandom.nextBoolean() ) {
			style |= OpponentCar.STYLE_DRIVES_RIGHT;
		}
		mOpponentCar.setOnStart( mTrack, style );
		//
		mOffRoadTimer = 3 * (1000 / (int)REFRESH_PERIOD);// 3 sec before set on road
		// to test Drawbridge animation
		//for( int i = 0; i < (int)(3000 / GameView.REFRESH_PERIOD); i++ ) mTrack.mAnimation.move( 0, 0 );
	}

	@Override
	public void run() {
		//System.out.println("GameLoop thread started. mRunning = " + mRunning);
		createOffscreenBuffer();
		mG.setColor( Color.BLACK );
		mG.fillRect( 0, 0, SCREEN_WIDTH, SCREEN_HEIGHT );
		repaint();
		mAlpha = 0.0f;
		renderSplash( mG );
		//
		while( mIsRunning ) {
			final long curr_time = System.currentTimeMillis();
synchronized( mPaintLock ){
			try{
				if( mMode == MODE_RACE ) {
					if( mIsRace & ! mIsPractice ) Car.interaction( mTrack, mCar, mOpponentCar );
					//
					int ret = mCar.move( mTrack, curr_time );
					if( ret > 0 ) {
						Dashboard.freezeLapTime( --ret > 0 );// 3 sec to show lap time
					}
					mOpponentCar.move( mTrack, curr_time );
					mCar.calculateCameraCoordinates();
					//mOpponentCar.calculateCameraCoordinates();
					// decoration
					Render.drawHorizon( mG, mCamera );
					Render.drawEnvironment( mG, mCamera, mScenery );
					// track and opponent car
					mOpponentCar.mIsRendered = mIsPractice;
					Render.drawTrack( mG, mCamera, mOpponentCar );
					//
					if( mCar.mIsOffRoad & mCar.mIsTouchingRoad ) {
						Dashboard.initClouds();
						Dashboard.renderClouds( mG );
						if( --mOffRoadTimer <= 0 ) {
							mOffRoadTimer = (int)(3 * 1000 / REFRESH_PERIOD);// 3 sec before set on road
							Dashboard.stopAnimation();
							mCar.hookOnCrane( mTrack.mPieces );
						}
					}
					Dashboard.render( mG, mCar, mKey, mCar.mTravelDistance > mOpponentCar.mTravelDistance );
					//
					char[] text = null;
					if( mCar.mIsOnChains ) {
						if( mCar.mIsOnRoad ) {
							text = strDrop;
						}
						mIsRace |= mCar.moveCrane();
					}
					mCar.carBehaviour( mTrack, mKey );
					// to test code
						//Render.drawOpponentCar( mG, mOpponentCar, false );
						//mCar.lookForPoint( mTrack, mOpponentCar );
						//mCar.lookOnPoint( mTrack, 11, 6 );
						//mCar.moveCar( mKey );
					// to view opponent car model
						//mOpponentCar.rotate();
					if( mIsRace & ! mIsPractice )
						mOpponentCar.carBehaviour( mTrack );
					//
					if( mTrack.mAnimation != null ) mTrack.mAnimation.move( mCar.mPieceIndex, mOpponentCar.mPieceIndex );
					//
					if( mCar.mDamage >= Car.DAMAGE_LIMIT ) {
						text = strWreck;
						if( --mOffRoadTimer <= 0 ) {
							mMode = MODE_WAIT_KEY;
							mIsRaceWon = false;
							renderKeyRequest( mG );
							text = null;// to prevent from text overlap
						}
					}
					//
					if( mCar.mLap > MAX_LAPS || mOpponentCar.mLap > MAX_LAPS ) {// race is finished
						text = strLapsOver;
						if( ! mIsPractice ) {
							if( ! mIsRaceFinished ) {
								mIsRaceWon = mCar.mTravelDistance > mOpponentCar.mTravelDistance;
								mIsRaceFinished = true;
							}
							text = mIsRaceWon ? strRaceWon : strRaceLost;
						}
						if( --mOffRoadTimer <= 0 ) {
							mMode = MODE_WAIT_KEY;
							renderKeyRequest( mG );
							text = null;// to prevent from text overlap
						}
					}
					//
					if( text != null ) {
						Color color = Color.BLACK;
						if( ((mFrameCounter / (1000 / (int)REFRESH_PERIOD / 2)) & 1) != 0 )// 2 Hz
							color = Color.WHITE;
						mG.setColor( color );
						mG.drawChars( text, 0, text.length, (SCREEN_WIDTH - text.length * FONT_SYMBOL_WIDTH) >> 1, 25 * PF + FONT_BASELINE );
					}
					//
					mFrameCounter++;
				} else if( mMode == MODE_NAME ) {
					char typed_key = mKeyTyped;
					if( typed_key != '\0' ) {
						if( renderNameRequest( mG, typed_key ) ) {
							renderDivisions( mG );
							mMode = MODE_DIVISIONS;
							mAlpha = -1.0f;
						}
						mKeyTyped = '\0';// ready to next symbol
						mPrevKey = 0;
					}
				} else if( mMode == MODE_WAIT_KEY ) {// decelerate engine sound
					Dashboard.freezeLapTime( false );// stop lap timer
					mCar.mSpeed_fp8 = 0;// stop wheels
					mCar.deccelerateEngine();
					// --- update screen to finish animation
					mCar.calculateZXYmatrix();
					mCar.calculateCameraCoordinates();
					// decoration
					Render.drawHorizon( mG, mCamera );
					Render.drawEnvironment( mG, mCamera, mScenery );
					// track and opponent car
					mOpponentCar.mIsRendered = mIsPractice;
					Render.drawTrack( mG, mCamera, mOpponentCar );
					Dashboard.renderClouds( mG );
					Dashboard.render( mG, mCar, mKey, mCar.mTravelDistance > mOpponentCar.mTravelDistance );
					renderKeyRequest( mG );
					if( mKey != mPrevKey ) {
						mPrevKey = mKey;
						if( (mPrevKey & (KEY_FIRE | KEY_UP)) != 0 ) {
							mAlpha = -1.0f;
							SoundManager.stopAll();// turn off sound
							mTrack.clear();
							Render.freeTrack();
							mG.setClip( null );// restore clip area
							if( mIsPractice ) {
								mMode = MODE_MENU;
								renderMenu( mG, false, false );
							} else {
								mMode = MODE_SHOW_WON_OR_LOST;
								String picture = "lost";
								if( mCar.mDamage >= Car.DAMAGE_LIMIT ) {
									picture = "wreck";
								}
								if( mIsRaceWon ) {
									picture = "won";
								}
								renderBackgroundImage( mG, picture );
							}
						}
					}
				} else if( mKey != mPrevKey ) {
					mPrevKey = mKey;
					if( mKey != 0 ) {
						boolean isFire = (mPrevKey & KEY_FIRE) != 0;
						if( isFire ) {
							mAlpha = -1.0f;
						}
						switch( mMode ) {
						case MODE_MENU:
							if( isFire ) {
								mAlpha = 1.0f;
								renderMenu( mG, mMenu == menuMain, true );
								repaint();
								mPaintLock.wait( 300L );
								if( mMenu == menuMain ) {
									/*if( mMenu[mMenuSelectedItem] == strExit ) {
										((GameForm)getParent()).exit();
										break;
									}*/
									if( mMenu[mMenuSelectedItem] == strHallOfFame ) {
										mMode = MODE_HALL_OF_FAME;
										mAlpha = -1.0f;
										renderHallOfFame( mG );
										break;// case MODE_MENU
									} else if( mMenu[mMenuSelectedItem] == strPractise ) {
										mMenuSelectedItem = mCurrentDivision;
										mMenu = menuPractise;
										mIsPractice = true;
									} else if( mMenu[mMenuSelectedItem] == strLoadSaveReplay) {
										mMenuSelectedItem = 3;// Cancel
										mMenu = menuLoadSaveReplay;
									}
									else if( mMenu[mMenuSelectedItem] == strStartTheRacingSeason ) {
										// save current division states to be able do replay
										saveState();
										// set player position in the division at the last position
										final Division division = mDivisions[mCurrentDivision];
										for( int d = 0; d < DRIVERS_IN_DIVISION - 1; d++ ) {
											if( mPlayer == division.mDrivers[d] ) {
												division.mDrivers[d] = division.mDrivers[d + 1];
												division.mDrivers[d + 1] = mPlayer;
											}
										}
										// generate results for all divisions
										mPlayer.resetState();
										for( int d = 0; d < DIVISION_COUNT; d++ ) {
											generateRaceResult( d, 0, 1 );
											if( d != mCurrentDivision ) {
												generateRaceResult( d, 0, 2 );
												generateRaceResult( d, 1, 2 );
											}
										}
										//
										mCurrentOpponent = 0;
										mCurrentTrack = TRACKS_IN_DIVISION - 1;
										initTrack( mDivisions[mCurrentDivision].mTracks[mCurrentTrack] );
										//
										mAlpha = -1.0f;
										renderRaceOpponents( mG );
										mMode = MODE_RACE_OPPONENTS;
										mIsPractice = false;
										break;// case MODE_MENU
									}
								} else if( mMenu == menuPractise ) {
									if( mMenu[mMenuSelectedItem] == strTracksInDivision1 ) {
										mMenu = menuTrackSelectDiv1;
									} else if( mMenu[mMenuSelectedItem] == strTracksInDivision2 ) {
										mMenu = menuTrackSelectDiv2;
									} else if( mMenu[mMenuSelectedItem] == strTracksInDivision3 ) {
										mMenu = menuTrackSelectDiv3;
									} else if( mMenu[mMenuSelectedItem] == strTracksInDivision4 ) {
										mMenu = menuTrackSelectDiv4;
									}
									mMenuSelectedItem = 0;
								} else if( mMenu == menuTrackSelectDiv1 || mMenu == menuTrackSelectDiv2 ||
										mMenu == menuTrackSelectDiv3 || mMenu == menuTrackSelectDiv4 ) {
									if( mMenu[mMenuSelectedItem] == strCancel ) {
										mMenu = menuMain;
										mMenuSelectedItem = 1;// Practice
									} else {
										if( mMenu[mMenuSelectedItem] == strPractiseTheSkiJump ) {
											initTrack( mDivisions[0].mTracks[1] );
										} else if( mMenu[mMenuSelectedItem] == strPractiseTheDrawbridge ) {
											initTrack( mDivisions[0].mTracks[0] );
										} else if( mMenu[mMenuSelectedItem] == strPractiseTheHighJump ) {
											initTrack( mDivisions[1].mTracks[1] );
										} else if( mMenu[mMenuSelectedItem] == strPractiseTheRollerCoaster ) {
											initTrack( mDivisions[1].mTracks[0] );
										} else if( mMenu[mMenuSelectedItem] == strPractiseTheSteppingStones ) {
											initTrack( mDivisions[2].mTracks[1] );
										} else if( mMenu[mMenuSelectedItem] == strPractiseTheBigRamp ) {
											initTrack( mDivisions[2].mTracks[0] );
										} else if( mMenu[mMenuSelectedItem] == strPractiseTheLittleRamp ) {
											initTrack( mDivisions[3].mTracks[1] );
										} else if( mMenu[mMenuSelectedItem] == strPractiseTheHumpBack ) {
											initTrack( mDivisions[3].mTracks[0] );
										}
										mAlpha = -1.0f;
										mMode = MODE_TRACK_PREVIEW;
										renderTrackPreview( mG, mTrack.mName );
										break;// case MODE_MENU
									}
								} else if( mMenu == menuLoadSaveReplay ) {
									if( mMenu[mMenuSelectedItem] == strLoad ) {
										final JFileChooser fc = new JFileChooser();
										fc.setSelectedFile( new File( APP_NAME + ".log" ) );
										int state = fc.showOpenDialog( this );
										if( state == JFileChooser.APPROVE_OPTION ) {
											if( load( fc.getSelectedFile() ) < 0 ) {
												System.out.println("File load error for: " + fc.getSelectedFile() );
											}
										}
										renderDivisions( mG );
										mMode = MODE_DIVISIONS;
										break;// case MODE_MENU
									} else if( mMenu[mMenuSelectedItem] == strSave ) {
										final JFileChooser fc = new JFileChooser();
										fc.setSelectedFile( new File( APP_NAME + ".log" ) );
										int state = fc.showSaveDialog( this );
										if( state == JFileChooser.APPROVE_OPTION ) {
											save( fc.getSelectedFile() );
										}
									} else if( mMenu[mMenuSelectedItem] == strReplay ) {
										// restore stored state
										for( int d = 0; d < DIVISION_COUNT; d++ ) {
											System.arraycopy( mBackupDivisionState[d], 0, mDivisions[d].mDrivers, 0, DRIVERS_IN_DIVISION );
										}
										renderDivisions( mG );
										mMode = MODE_DIVISIONS;
										break;// case MODE_MENU
									} else if( mMenu[mMenuSelectedItem] == strCancel ) {
										mMenu = menuMain;
										mMenuSelectedItem = 1;// Practice
									}
								}// if menu
							}// if KEY_FIRE
							if( (mPrevKey & KEY_UP) != 0 ) {
								if( --mMenuSelectedItem < 0 ) { mMenuSelectedItem++; }
							} else if( (mPrevKey & KEY_DOWN) != 0 ) {
								if( ++mMenuSelectedItem >= mMenu.length ) { mMenuSelectedItem--; }
							}
							renderMenu( mG, mMenu == menuMain, false );
							break;// case MODE_MENU
						case MODE_TRACK_PREVIEW:
							if( isFire ) {
								initRace();
								Dashboard.renderPillars( mG );
								mMode = MODE_RACE;
								break;// case MODE_TRACK_PREVIEW
							}
							if( (mPrevKey & KEY_RIGHT) != 0 ) {
								mCamera.rotate( 1, Track.TRACKS_CENTER_X, Track.TRACKS_CENTER_Y );
							} else if( (mPrevKey & KEY_LEFT) != 0 ) {
								mCamera.rotate( -1, Track.TRACKS_CENTER_X, Track.TRACKS_CENTER_Y );
							}
							renderTrackPreview( mG, mTrack.mName );
							break;// case MODE_TRACK_PREVIEW
						case MODE_RACE_OPPONENTS:
							if( isFire ) {
								mMode = MODE_TRACK_PREVIEW;
								renderTrackPreview( mG, mTrack.mName );
							}
							break;// case MODE_RACE_OPPONENTS
						case MODE_HALL_OF_FAME:
							if( isFire ) {
								mMenu = menuMain;
								mMenuSelectedItem = 1;
								mMode = MODE_MENU;
								renderMenu( mG, true, false );
							}
							break;// case MODE_HALL_OF_FAME
						case MODE_DIVISIONS:
							if( isFire ) {
								mMenu = menuMain;
								mMenuSelectedItem = 1;
								mMode = MODE_MENU;
								renderMenu( mG, true, false );
							}
							break;// case MODE_DIVISIONS
						case MODE_PROMOTION_PICTURE:
							if( isFire ) {
								renderDivisions( mG );
								mMode = MODE_DIVISIONS;
							}
							break;
						case MODE_PROMOTION:
							if( isFire ) {
								renderBackgroundImage( mG, "promotion" );
								mMode = MODE_PROMOTION_PICTURE;
							}
							break;// case MODE_PROMOTION
						case MODE_RACE_RESULT_STATE:
							if( isFire ) {
								if( --mCurrentTrack < 0 ) {
									mCurrentTrack = TRACKS_IN_DIVISION - 1;
									if( ++mCurrentOpponent > (DRIVERS_IN_DIVISION - 1 - 1) ) {
										mCurrentOpponent = 0;
										// sorting
										for( int i = 0; i < DIVISION_COUNT; i++ ) {
											sortDivision( i );
										}
										//
										renderPromotion( mG );
										mMode = MODE_PROMOTION_PICTURE;
										// check player state
										if( mPlayer == mDivisions[mCurrentDivision].mDrivers[0] ) {
											mMode = MODE_PROMOTION;// promotion
											if( mCurrentDivision > 0 ) mCurrentDivision--;
										}
										if( mPlayer == mDivisions[mCurrentDivision].mDrivers[DRIVERS_IN_DIVISION - 1] ) {
											mMode = MODE_PROMOTION_PICTURE;// relegation
											if( mCurrentDivision < (DIVISION_COUNT - 1) ) mCurrentDivision++;
										}
										// promotion on divisions
										for( int d = 0; d < DIVISION_COUNT - 1; d++ ) {
											Driver driver = mDivisions[d].mDrivers[DRIVERS_IN_DIVISION - 1];
											mDivisions[d].mDrivers[DRIVERS_IN_DIVISION - 1] = mDivisions[d + 1].mDrivers[0];
											mDivisions[d + 1].mDrivers[0] = driver;
										}
										break;// case MODE_RACE_RESULT_POINTS
									}
								}
								initTrack( mDivisions[mCurrentDivision].mTracks[mCurrentTrack] );
								//
								renderRaceOpponents( mG );
								mMode = MODE_RACE_OPPONENTS;
							}
							break;// case MODE_RACE_RESULT_POINTS
						case MODE_RACE_RESULT_POINTS:
							if( isFire ) {
								mMode = MODE_RACE_RESULT_STATE;
								renderRaceResultState( mG );
							}
							break;// case MODE_RACE_RESULT_POINTS
						case MODE_RACE_RESULT:
							if( isFire ) {
								mMode = MODE_RACE_RESULT_POINTS;
								renderRaceResultPoints( mG );// and calculating
							}
							break;// case MODE_RACE_RESULT
						case MODE_SHOW_WON_OR_LOST:
							if( isFire ) {
								if( (mCar.mBestLapTime / 10 < mTrack.mBestLapTimeMs10) ||
									(mCar.mRaceTime / 10 < mTrack.mBestRaceTimeMs10) )
								{
									mMode = MODE_RACE_RESULT;
									renderRaceResult( mG, mCar );
									break;
								}
								mMode = MODE_RACE_RESULT_POINTS;
								renderRaceResultPoints( mG );// and calculating
							}
							break;// case MODE_SHOW_WON_OR_LOST
						case MODE_SPLASH:
							if( isFire ) {
								mMode = MODE_NAME;
								mKeyTyped = '\0';// clear typed key to prevent incorrect first symbol
								// mKeyTyped == ' ', because FIRE button is space
								renderNameRequest( mG, ' ' );
								break;// case MODE_SPLASH
							}
							if( (mPrevKey & (KEY_RIGHT | KEY_DOWN)) != 0 ) {
								if( ++mMenuSelectedItem >= LANGUAGES.length ) mMenuSelectedItem = 0;
							} else if( (mPrevKey & (KEY_LEFT | KEY_UP)) != 0 ) {
								if( --mMenuSelectedItem < 0 ) mMenuSelectedItem += LANGUAGES.length;
							}
							setLanguage( LANGUAGES[mMenuSelectedItem] );
							renderSplash( mG );
							break;// case MODE_SPLASH
						}// switch( mMode )
					}// if( mKey != 0 )
				}// if( mKey != mPrevKey )
			} catch(Exception e) {
				System.out.println( e.getMessage() );
				e.printStackTrace();
			}
			finally {
			}
}// synchronized
			waiting( curr_time );
		}
	}
}