import java.awt.Color;

class OpponentCar extends Car {
	// driving styles
	static final int STYLE_OBSTRUCTS_PLAYER = 1;
	static final int STYLE_PUSH_PLAYER = 2;
	static final int STYLE_WHEELIE = 4;
	static final int STYLE_DRIVES_NEAR_EDGE = 8;
	static final int STYLE_DRIVES_RIGHT = 16;
	//
	private static final int POSITION_ACCURACY = 16;// cm
	//
	private final Color COLOR_FLOOR = new Color( 32, 32, 32 );
	private final Color COLOR_SIDE = new Color( 172, 32, 32 );
	private final Color COLOR_TOPSIDE = Color.DARK_GRAY;
	private final Color COLOR_ROOF = Color.DARK_GRAY;
	private final Color COLOR_WINDOW = new Color( 96, 128, 128 );
	private final Color COLOR_BONNET = COLOR_SIDE;
	private final Color COLOR_BUMPER = COLOR_SIDE;
	final Color COLOR_TYRE = new Color( 32, 32, 32 );
	final Color COLOR_WHEEL = new Color( 16, 16, 16 );
	final Color COLOR_DISK = new Color( 128, 128, 128 );
	/** Polygon coordinates indexes must be in counter clock wise order */
	final Model mHull = new Model( new Coord3D[] {
			// floor
			new Coord3D( -(WIDTH / 2 - WHEEL_WIDTH),  LENGTH / 2/* - LENGTH / 16*/, WHEEL_RADIUS - 8 ),// 0
			new Coord3D(  (WIDTH / 2 - WHEEL_WIDTH),  LENGTH / 2/* - LENGTH / 16*/, WHEEL_RADIUS - 8 ),// 1
			new Coord3D(  (WIDTH / 2 - WHEEL_WIDTH), -LENGTH / 2, WHEEL_RADIUS - 8 ),// 2
			new Coord3D( -(WIDTH / 2 - WHEEL_WIDTH), -LENGTH / 2, WHEEL_RADIUS - 8 ),// 3
			// wing and door tops
			new Coord3D( -(WIDTH / 2 - WHEEL_WIDTH), LENGTH / 2/* - LENGTH / 16*/, 64 - 8 ),// 4
			new Coord3D(  (WIDTH / 2 - WHEEL_WIDTH), LENGTH / 2/* - LENGTH / 16*/, 64 - 8 ),// 5
			//
			new Coord3D( -(WIDTH / 2 - WHEEL_WIDTH), LENGTH / 4, HEIGHT - 40 ),// 6
			new Coord3D(  (WIDTH / 2 - WHEEL_WIDTH), LENGTH / 4, HEIGHT - 40 ),// 7
			//
			new Coord3D( -(WIDTH / 2 - WHEEL_WIDTH), - LENGTH / 8, HEIGHT - 40 ),// 8
			new Coord3D(  (WIDTH / 2 - WHEEL_WIDTH), - LENGTH / 8, HEIGHT - 40 ),// 9
			//
			new Coord3D( -(WIDTH / 2 - WHEEL_WIDTH), -LENGTH / 2 + LENGTH / 16, HEIGHT - 40 ),// 10
			new Coord3D(  (WIDTH / 2 - WHEEL_WIDTH), -LENGTH / 2 + LENGTH / 16, HEIGHT - 40 ),// 11
			// roof
			new Coord3D( -(WIDTH / 2 - WHEEL_WIDTH - WIDTH / 8), LENGTH / 32, HEIGHT ),// 12
			new Coord3D(  (WIDTH / 2 - WHEEL_WIDTH - WIDTH / 8), LENGTH / 32, HEIGHT ),// 13
			//
			new Coord3D( -(WIDTH / 2 - WHEEL_WIDTH - WIDTH / 8), - LENGTH / 8, HEIGHT ),// 14
			new Coord3D(  (WIDTH / 2 - WHEEL_WIDTH - WIDTH / 8), - LENGTH / 8, HEIGHT ),// 15
			//
			new Coord3D( -(WIDTH / 2 - WHEEL_WIDTH - WIDTH / 8), -LENGTH / 4, HEIGHT ),// 16
			new Coord3D(  (WIDTH / 2 - WHEEL_WIDTH - WIDTH / 8), -LENGTH / 4, HEIGHT ),// 17
			// for smooth bonnet
			//new Coord3D( -(WIDTH / 2 - WHEEL_WIDTH - 16), LENGTH / 2/* - LENGTH / 16*/, 64 ),// 18
			//new Coord3D(  (WIDTH / 2 - WHEEL_WIDTH - 16), LENGTH / 2/* - LENGTH / 16*/, 64 ),// 19
		}, new Polygon[] {
			// this order is important, if we see front window, z-order is window-bonnet-bumper
			// front window
			new Polygon( COLOR_WINDOW, new int[] { 7, 6, 12, 13 } ),
			// --- simple
			// bonnet
			new Polygon( COLOR_BONNET, new int[] { 5, 4, 6, 7 } ),
			// bumper
			new Polygon( COLOR_BUMPER, new int[] { 1, 0, 4, 5 } ),
			// --- end simple
			// --- start smooth
			//new Polygon( COLOR_BONNET, new int[] { 19, 18, 6, 7 } ),
			//new Polygon( COLOR_BONNET, new int[] { 19, 7, 5 } ),
			//new Polygon( COLOR_BONNET, new int[] { 18, 4, 6 } ),
			// bumper
			//new Polygon( COLOR_BUMPER, new int[] { 1, 0, 4, 18, 19, 5 } ),
			// --- end smooth
			// if we see rear window, z-order is side-rear window
			// left side1
			new Polygon( COLOR_SIDE, new int[] { 0, 3, 10, 8, 6, 4 } ),
			// left side2, area behind side windows
			new Polygon( COLOR_TOPSIDE, new int[] { 8, 10, 16, 14 } ),
			// right side1
			new Polygon( COLOR_SIDE, new int[] { 2, 1, 5, 7, 9, 11 } ),
			// right side2, area behind side windows
			new Polygon( COLOR_TOPSIDE, new int[] { 11, 9, 15, 17 } ),
			// left window
			new Polygon( COLOR_WINDOW, new int[] { 6, 8, 14, 12 } ),
			// right window
			new Polygon( COLOR_WINDOW, new int[] { 9, 7, 13, 15 } ),
			// rear
			new Polygon( COLOR_SIDE, new int[] { 3, 2, 11, 10 } ),
			// rear window
			new Polygon( COLOR_WINDOW, new int[] { 10, 11, 17, 16 } ),
			// if we see roof, z-order is bonnet-side-window-roof or floor.
			// we see either roof or floor
			// roof
			new Polygon( COLOR_ROOF, new int[] { 13, 12, 14, 16, 17, 15 } ),
			// floor
			new Polygon( COLOR_FLOOR, new int[] { 0, 1, 2, 3 } )
		}
	).shift( 0, -CG_Y, SPRINT_TRAVEL / 2 - CG_Z );
	//
	static final int WHEEL_CIRCLE_POINTS = 12;// must be even
	/** y coordinates for wheel circle */
	final int[] mWheelY;
	/** z coordinates for wheel circle */
	final int[] mWheelZ;
	/** point indexes to check visibility of the left side */
	final int left_side_0 = 0;
	final int left_side_1 = 3;
	final int left_side_2 = 10;
	/** point indexes to check visibility of the right side */
	final int right_side_0 = 2;
	final int right_side_1 = 1;
	final int right_side_2 = 5;
	/** true if car was rendered on current frame */
	boolean mIsRendered = false;
	/** track's max speed */
	private int mMaxSpeed_fp8 = 0;
	/** driving style */
	private int mStyle = 0;

	OpponentCar() {
		// calculate wheels
		mWheelY = new int[WHEEL_CIRCLE_POINTS];
		mWheelZ = new int[WHEEL_CIRCLE_POINTS];
		for( int i = 0; i < WHEEL_CIRCLE_POINTS; i++ ) {
			int angle = i * Camera.ANGLE_360 / WHEEL_CIRCLE_POINTS;
			angle = ((angle > Camera.ANGLE_180) ? (angle - Camera.ANGLE_360) : angle);
			int abs_angle = Math.abs( angle );
			int sin;
			int cos;
			if( abs_angle <= Camera.ANGLE_90 ) {
				cos = (int)Camera.sCosTable[abs_angle];
				sin = (int)Camera.sCosTable[Camera.ANGLE_90 - abs_angle];
			} else {
				cos = -(int)Camera.sCosTable[Camera.ANGLE_180 - abs_angle];
				sin = (int)Camera.sCosTable[abs_angle - Camera.ANGLE_90];
			}
			if( angle < 0 ) {
				sin = -sin;
			}
			// sin
			mWheelZ[i] = WHEEL_RADIUS * sin / Camera.PRECISION + (WHEEL_RADIUS);
			// cos
			mWheelY[i] = WHEEL_RADIUS * cos / Camera.PRECISION;
		}
	}
	/*
	private boolean mIsUp = true;
	void rotate() {
		setZangle( z_angle + 1 * Camera.ANGLE_360 / 360 );
		//
		int bottom_z = mFrontLeftWheelBottom;
		if( mIsUp ) {
			bottom_z++;
			if( bottom_z >= WHEEL_BOTTOM_Z + SPRINT_TRAVEL ) {
				mIsUp = false;
			}
		} else {
			bottom_z--;
			if( bottom_z <= WHEEL_BOTTOM_Z ) {
				mIsUp = true;
			}
		}
		mFrontLeftWheelBottom = bottom_z;
		mFrontRightWheelBottom = bottom_z;
		mRearLeftWheelBottom = bottom_z;
		mRearRightWheelBottom = bottom_z;
	}
	*/
	/**
	 * Set initial parameters for opponent car
	 * @param track
	 * @param style
	 */
	void setOnStart(final Track track, int style) {
		mLap = 0;
		mBestLapTime = 0x7fffffff;
		mRaceTime = 0x7fffffff;
		//
		final int distance_to_start = 4;
		mLap = 0;
		mPieceIndex = track.mPieces.length;
		TrackPiece tp = track.mPieces[--mPieceIndex];
		mPieceCoordIndex = tp.mCoords.length - (distance_to_start + 1) * 2;
		mCoordIndex = track.mNumTrackCoords - distance_to_start * 2;
		/*
		mPieceIndex = 0;
		tp = track.mPieces[mPieceIndex];
		mPieceCoordIndex = tp.mCoords.length - (4 + 1) * 2;
		mCoordIndex = mPieceCoordIndex;
		for( int i = 0; i < mPieceIndex; i++ ) {
			mCoordIndex += track.mPieces[i].mCoords.length;
			mCoordIndex -= 2;
		}
		*/
		mTravelDistance = 0;
		mDamage = 0;
		mStartRaceTime = 0;
		mBestLapTime = 0x7fffffff;
		mRaceTime = 0x7fffffff;
		mBoostReserve = track.mStandardBoost;
		mBoostFrameCounter = 0;
		mIsBoostActivated = false;
		// reset speeds
		mRoadToFrontLeftWheelDistance_fp = 0;
		mRoadToFrontRightWheelDistance_fp = 0;
		mRoadToRearLeftWheelDistance_fp = 0;
		mRoadToRearRightWheelDistance_fp = 0;
		mRotationSpeedX = 0;
		mRotationSpeedY = 0;
		mRotationSpeedZ = 0;
		mSpeedX_fp8 = 0;
		mSpeedY_fp8 = 0;
		mSpeedZ_fp8 = 0;
		mSpeed_fp8 = 0;
		mIsAccelerating = false;
		//
		mMaxSpeed_fp8 = track.mMaxSpeed_fp8 + GameView.mRandom.nextInt( track.mSpeedDeviation_fp8 );
		//mMaxSpeed_fp8 = 500 << 8;// no limit to test
		//mMaxSpeed_fp8 = track.mMaxSpeed_fp8;// minimal speed to test
		//mMaxSpeed_fp8 = track.mMaxSpeed_fp8 + track.mSpeedDeviation_fp8;// maximal speed to test
		mStyle = style;
		//
		//mStyle = STYLE_DRIVES_NEAR_EDGE;
		//mStyle |= STYLE_DRIVES_RIGHT;
		//mStyle |= STYLE_OBSTRUCTS_PLAYER;
		//mStyle |= STYLE_PUSH_PLAYER;
		//mStyle |= STYLE_WHEELIE;
		//
		setAngles( 0, 0, Track.getSectionAngle( tp, mPieceCoordIndex ) );
		calculateZXYmatrix();
		//
		final int mr = 3;
		final int shift = -(WIDTH / 2 + CENTER_LINE_MARGIN);
		Coord3D cleft = tp.mCoords[mPieceCoordIndex];
		Coord3D cright = tp.mCoords[mPieceCoordIndex + 1];
		int x = (cleft.mX + cright.mX) >> 1;
		int y = (cleft.mY + cright.mY) >> 1;
		int z = Math.max( cleft.mZ, cright.mZ );
		z += CG_Z;
		x += ((mMatrix[0 * mr + 0] * shift) >> Camera.PRECISION_LOG2);
		y += ((mMatrix[0 * mr + 1] * shift) >> Camera.PRECISION_LOG2);
		setPosition( x, y, z );
	}
	void carBehaviour(final Track track) {
		final int mr = 3;
		final int[] m = mMatrix;
		//--------------------------------------------------------------------------
		// calculate forward speed
		mSpeed_fp8 = (mSpeedX_fp8 * m[1 * mr + 0] + mSpeedY_fp8 * m[1 * mr + 1] + mSpeedZ_fp8 * m[1 * mr + 2]) >> (Camera.PRECISION_LOG2);
		//--------------------------------------------------------------------------
		int x_rotation_acceleration = 0;
		int y_rotation_acceleration = 0;
		//int z_rotation_acceleration = 0;
		//--------------------------------------------------------------------------
		// calculate accelerations
		int x_acceleration_fp8 = 0;
		int y_acceleration_fp8 = 0;
		int z_acceleration_fp8 = 0;
		//--------------------------------------------------------------------------
		//final int slope_factor_fp = WHEEL_RADIUS * track.getSectionSlopeCos( mCoordIndex );
		final int average_acceleration_fp8 = calculateWheelsStroke( track );
		if( mIsTouchingRoad ) {
			/*
			cos = sqrt( 1*1 - sin*sin ) = mSinToCos[sin]
			angles in range -90 ... +90 degrees
			z_inclination_to_road is zero because road exists in X and Y planes only
			calculate ZXY matrix to convert from world to local coordinates
			x = x * cos z * cos y + x * sin z * sin x * sin y + y * cos z * sin x * sin y - y * sin z * cos y + z * cos x * sin y
			y = x * sin z * cos x + y * cos z * cos x - z * sin x
			z = x * sin z * sin x * cos y - x * cos z * sin y + y * cos z * sin x * cos y + y * sin z * sin y + z * cos x * cos y
			we have only z value, so
			x = z * cos x * sin y
			y = - z * sin x = z * (- sin x)
			z = z * cos x * cos y
			To simplify, calculate sin x with reverse sign */
			//--------------------------------------------------------------------------
			//
			final int longtitude_difference_fp8 = (mRoadToRearLeftWheelDistance_fp + mRoadToRearRightWheelDistance_fp -
					mRoadToFrontLeftWheelDistance_fp - mRoadToFrontRightWheelDistance_fp) >> (1 + Camera.PRECISION_LOG2 - 8);
			//
			//final int longtitude_difference_fp8 =
			//	(Math.max( mRoadToRearLeftWheelDistance_fp, mRoadToRearRightWheelDistance_fp ) -
			//	Math.max( mRoadToFrontLeftWheelDistance_fp, mRoadToFrontRightWheelDistance_fp) ) >> (Camera.PRECISION_LOG2 - 8);
			//
			/* Calculate x inclination to road, sin x with reverse sign
			tan x = (df - dr) / length
			but for small angle can use:
			sin x = (df - dr) / length
			and need to limit result to 1 */
			final int surface_neg_sinx_fp8 = Math.min( Math.max( (longtitude_difference_fp8) / WHEEL_FRONT_REAR_DISTANCE, -256 ), 256 );
			// get cosines
			final int surface_cosx_fp8 = (int)mSinToCosTable_fp8[Math.abs( surface_neg_sinx_fp8 )];
			/*
			final int lateral_difference_fp8 = (mRoadToFrontLeftWheelDistance_fp + mRoadToRearLeftWheelDistance_fp -
					mRoadToFrontRightWheelDistance_fp - mRoadToRearRightWheelDistance_fp) >> (1 + Camera.PRECISION_LOG2 - 8);
			*/
			final int lateral_difference_fp8 =
				(Math.max( mRoadToFrontLeftWheelDistance_fp, mRoadToRearLeftWheelDistance_fp) -
				Math.max( mRoadToFrontRightWheelDistance_fp, mRoadToRearRightWheelDistance_fp ) ) >> (Camera.PRECISION_LOG2 - 8);
			//
			/* Calculate y inclination to road
			tan y = (dfl - dfr) / length
			but for small angle can use:
			sin x = (dfl - dfr) / length
			and need to limit result to 1 */
			final int surface_siny_fp8 = Math.min( Math.max( (lateral_difference_fp8) / WHEEL_TREAD, -256 ), 256 );
			// get cosines
			final int surface_cosy_fp8 = (int)mSinToCosTable_fp8[Math.abs( surface_siny_fp8 )];
			//--------------------------------------------------------------------------
			// car_road_collision_x_acceleration
			int car_road_x_acceleration_fp8 = (average_acceleration_fp8 * ((surface_cosx_fp8 * surface_siny_fp8) >> 8)) >> (8);
			// car_road_collision_y_acceleration
			final int car_road_y_acceleration_fp8 = (average_acceleration_fp8 * surface_neg_sinx_fp8) >> (8);
			// car_road_collision_z_acceleration
			final int car_road_z_acceleration_fp8 = (average_acceleration_fp8 * ((surface_cosx_fp8 * surface_cosy_fp8) >> 8)) >> (8);
			// car_road_z_acceleration always positive
			//final int double_car_road_z_acceleration_fp8 = (car_road_z_acceleration_fp8 << 1);
			//--------------------------------------------------------------------------
			// x and y accelerations is forward moving
			// x acceleration
			/*if( Math.abs( car_road_x_acceleration_fp8 ) > double_car_road_z_acceleration_fp8 ) {
				int x_accel = double_car_road_z_acceleration_fp8;
				if( car_road_x_acceleration_fp8 < 0 )
					x_accel = -x_accel;
				car_road_x_acceleration_fp8 = x_accel;
			}*/
			final int x_speed_fp8 = (mSpeedX_fp8 * m[0 * mr + 0] + mSpeedY_fp8 * m[0 * mr + 1] + mSpeedZ_fp8 * m[0 * mr + 2]) >> (Camera.PRECISION_LOG2);
			car_road_x_acceleration_fp8 -= x_speed_fp8;
			x_acceleration_fp8 += car_road_x_acceleration_fp8;
			y_acceleration_fp8 += car_road_y_acceleration_fp8;
			z_acceleration_fp8 += car_road_z_acceleration_fp8;
			//--------------------------------------------------------------------------
			if( mIsOffRoad ) {//|| (mDamage > DAMAGE_LIMIT && ! mIsScratchSide) ) {// brake if wrecked or touching ground
				x_acceleration_fp8 -= x_speed_fp8;
				y_acceleration_fp8 -= mSpeed_fp8 >> 1;
				//z_acceleration_fp8 -= 0;
				x_rotation_acceleration -= mRotationSpeedX >> 4;
				y_rotation_acceleration -= mRotationSpeedY >> 4;
				//z_rotation_acceleration -= mRotationSpeedZ >> 1;
				//System.out.println("OffRoad, maxSpeed = " + mMaxSpeed_fp8);
			}// if is off road
			//
			//--------------------------------------------------------------------------
			//if( mRoadToFrontLeftWheelDistance_fp > 0 || mRoadToFrontRightWheelDistance_fp > 0 ) {// left or right wheel touching the road
				// calculate difference angle
				int look_ahead_piece = mPieceIndex;
				int look_ahead_index = mPieceCoordIndex + 1 * 2;
				if( look_ahead_index >= track.mPieces[look_ahead_piece].mCoords.length - 2 ) {
					look_ahead_index -= mPieceCoordIndex;
					look_ahead_index += 2;
					if( ++look_ahead_piece >= track.mPieces.length ) look_ahead_piece -= track.mPieces.length;
				}
				//
				int angle_difference = Track.getSectionAngle( track.mPieces[look_ahead_piece], look_ahead_index ) - z_angle;
				if( angle_difference > Camera.ANGLE_180 ) angle_difference -= Camera.ANGLE_360;
				if( angle_difference < -Camera.ANGLE_180) angle_difference += Camera.ANGLE_360;
				// smoothing
				mRoadToCarAngleDifference += ((angle_difference - mRoadToCarAngleDifference) >> 2);
				//mRoadToCarAngleDifference = (angle_difference + mRoadToCarAngleDifference) >> 1;
				z_angle += mRoadToCarAngleDifference;
				//z_angle = track.getSectionAngle( track.mPieces[look_ahead_piece], look_ahead_index );
				//mRotationSpeedZ = 0;
				final TrackPiece tp = track.mPieces[mPieceIndex];
				// steering to suggested road position
				// calculate center road position
				Coord3D cleft = tp.mCoords[mPieceCoordIndex];
				Coord3D cright = tp.mCoords[mPieceCoordIndex + 1];
				//int middle_x =
				//	m[0 * mr + 0] * (cleft.mX - mCG_X + cright.mX - mCG_X) +
				//	m[0 * mr + 1] * (cleft.mY - mCG_Y + cright.mY - mCG_Y) +
				//	m[0 * mr + 2] * (cleft.mZ - mCG_Z + cright.mZ - mCG_Z);
				// around z
				//x = x * cos z - y * sin z
				int middle_x = cos_z * (cleft.mX - mCG_X + cright.mX - mCG_X);
				middle_x += sin_z * (mCG_Y - cleft.mY + mCG_Y - cright.mY);
				middle_x >>= (Camera.PRECISION_LOG2 + 1);
				if( Math.abs( angle_difference ) < 1 * Camera.ANGLE_360 / 360 ) {
					int shift = 0;// 0 if drive to center, < 0 if drive to left, > 0 if drive to right
					if( Math.abs( sDistance ) < 4 * 2) {// if close to player
						// to opposite side to free road
						shift = -(WIDTH / 2 + CENTER_LINE_MARGIN);
						if( sDistanceX > 0 ) {// opponent to right
							shift = (WIDTH / 2 + CENTER_LINE_MARGIN);
						}
						if( (mStyle & STYLE_PUSH_PLAYER) != 0 && (sDistance >= -1 * 2 && sDistance <= 1 * 2) ) {
							shift = -shift;// push player
						}
						if( (mStyle & STYLE_OBSTRUCTS_PLAYER) != 0 && sDistance > 1 * 2 ) {
							shift = -sDistanceX;// set to player position
						}
					} else {// if far distance
						// check drive style
						if( (mStyle & STYLE_DRIVES_NEAR_EDGE) != 0 ) {
							shift = -(WIDTH / 2 + CENTER_LINE_MARGIN);
							if( (mStyle & STYLE_DRIVES_RIGHT) != 0 ) {
								shift = (WIDTH / 2 + CENTER_LINE_MARGIN);
							}
						}
					}
					if( (shift < 0 && (middle_x > 0 || middle_x < -POSITION_ACCURACY + WIDTH / 2 + CENTER_LINE_MARGIN) ) ||
						(shift > 0 && (middle_x < 0 || middle_x > POSITION_ACCURACY - WIDTH / 2 - CENTER_LINE_MARGIN) ) )
					{
						middle_x += shift;
					}
					if( shift == 0 && (Math.abs( middle_x ) < POSITION_ACCURACY) ) {
						middle_x = 0;
					}
					middle_x >>= 1;
				}// if along track
				final int horizontal_shift_fp8 = (middle_x * mSpeed_fp8) >> (8 + 2);
				mCG_X_fp8 += (horizontal_shift_fp8 * m[0 * mr + 0]) >> Camera.PRECISION_LOG2;
				mCG_Y_fp8 += (horizontal_shift_fp8 * m[0 * mr + 1]) >> Camera.PRECISION_LOG2;
				mCG_Z_fp8 += (horizontal_shift_fp8 * m[0 * mr + 2]) >> Camera.PRECISION_LOG2;
			//}// left or right wheel touching the road
			//--------------------------------------------------------------------------
			// speed control
			int key = 0;
			int speed_diff_fp8 = tp.mRecommendedSpeed_fp8;
			if( speed_diff_fp8 < 0 ) {
				speed_diff_fp8 = -speed_diff_fp8;
				speed_diff_fp8 = Math.min( speed_diff_fp8, mMaxSpeed_fp8 );
			}
			//
			speed_diff_fp8 -= (mSpeed_fp8);// >> 8);
			if( speed_diff_fp8 > 0 ) {// if lower
				key |= GameView.KEY_UP | GameView.KEY_FIRE;
				//if( tp.mRecommendedSpeed > 0 ) key |= GameView.KEY_FIRE;
			} else if( speed_diff_fp8 < 0 ) {// if higher
				key |= GameView.KEY_DOWN;
				if( speed_diff_fp8 < -((int)(20 * GameView.REFRESH_PERIOD) / 36) ) {// 20 km/h
					key |= GameView.KEY_FIRE;// if speed too high
				}
			}
			//
			boolean is_boost = false;
			if( (mSpeed_fp8 < MAX_SPEED_FP8) && (mDamage < DAMAGE_LIMIT) &&
				(mIsAccelerating || ((key & (GameView.KEY_UP | GameView.KEY_DOWN)) != 0)) )
			{
				if( (key & GameView.KEY_FIRE) != 0 ) {
					if( mBoostReserve > 0 ) {
						/*if( ++mBoostFrameCounter >= (1000 / (int)GameView.REFRESH_PERIOD) ) {// 1.0 сек
							mBoostFrameCounter = 0;
							mBoostReserve--;
						}*/
						is_boost = true;
					}
				}
				mIsAccelerating = (key & GameView.KEY_DOWN) == 0;// backward
			}// if accelerating or decelerating
			mIsBoostActivated = is_boost;
			//
			//System.out.println(mPieceIndex + ", speed = " + (mSpeed_fp8 >> 8) + ", rec = " + (tp.mRecommendedSpeed_fp8 >> 8) + ", diff = " + (speed_diff_fp8 >> 8) + ", key = " + Integer.toHexString( key ) );
			//--------------------------------------------------------------------------
			if( mIsAccelerating || ( (key & GameView.KEY_DOWN) != 0 ) ) {
				// engine acceleration depends from current forward speed
				int engine_acceleration_fp8 = -(mSpeed_fp8 >> 8);
				if( mRoadToFrontLeftWheelDistance_fp > 0 ) engine_acceleration_fp8 += ENGINE_ACCELERATION_FP8;
				if( mRoadToFrontRightWheelDistance_fp > 0 ) engine_acceleration_fp8 += ENGINE_ACCELERATION_FP8;
				if( mRoadToRearLeftWheelDistance_fp > 0 ) engine_acceleration_fp8 += ENGINE_ACCELERATION_FP8;
				if( mRoadToRearRightWheelDistance_fp > 0 ) engine_acceleration_fp8 += ENGINE_ACCELERATION_FP8;
				//
				if( mIsBoostActivated ) {
					engine_acceleration_fp8 <<= 1;
				}
				// car accelerate depends from grip
				//engine_acceleration_fp8 = Math.min( engine_acceleration_fp8, double_car_road_z_acceleration_fp8 );
				if( (key & GameView.KEY_DOWN) != 0 ) {// backward
					engine_acceleration_fp8 = -engine_acceleration_fp8;
				}
				y_acceleration_fp8 += engine_acceleration_fp8;
				// to pitch while accelerating
				if( (mStyle & STYLE_WHEELIE) != 0 && mSpeed_fp8 < ((int)(256 * 40 * GameView.REFRESH_PERIOD) / 36)) {
					x_rotation_acceleration -= (engine_acceleration_fp8 >> (8 - 5));
				}
			}
			//--------------------------------------------------------------------------
			// ration x/y is WHEEL_FRONT_REAR_DISTANCE / (2 * WHEEL_X) ~= 2
			x_rotation_acceleration += (mRearLeftWheelAcceleration_fp + mRearRightWheelAcceleration_fp -
				mFrontLeftWheelAcceleration_fp - mFrontRightWheelAcceleration_fp) >> (1 + Camera.PRECISION_LOG2 - 5);// 6 max
			//
			y_rotation_acceleration = (mFrontRightWheelAcceleration_fp + mRearRightWheelAcceleration_fp -
				mFrontLeftWheelAcceleration_fp - mRearLeftWheelAcceleration_fp) >> (1 + Camera.PRECISION_LOG2 - 7);// 9 max
		}// if mIsTouchingRoad
		// if on edge or fly do small forward rotation
		if( ! mIsTouchingRoad ) {// 17 deg/s^2
			x_rotation_acceleration = Math.max( x_rotation_acceleration, (int)(17 * Camera.ANGLE_360 / 360 * GameView.REFRESH_PERIOD * GameView.REFRESH_PERIOD / 1000 / 1000) );
		}
		x_rotation_acceleration -= (mRotationSpeedX >> 4);
		y_rotation_acceleration -= (mRotationSpeedY >> 4);
		//--------------------------------------------------------------------------
		// the effect of air resistance and road load
		// Check if slipstream
		if( sDistance < 0 && sDistance > -4 * 2) {// if close to player
			if( Math.abs( sDistanceX ) < WIDTH ) {
				y_acceleration_fp8 += mSpeed_fp8 >> 16;
			}
		}
		y_acceleration_fp8 -= mSpeed_fp8 >> 8;
		int world_x_acceleration_fp8 = (x_acceleration_fp8 * m[0 * mr + 0] + y_acceleration_fp8 * m[1 * mr + 0] + z_acceleration_fp8 * m[2 * mr + 0]) >> (Camera.PRECISION_LOG2);
		int world_y_acceleration_fp8 = (x_acceleration_fp8 * m[0 * mr + 1] + y_acceleration_fp8 * m[1 * mr + 1] + z_acceleration_fp8 * m[2 * mr + 1]) >> (Camera.PRECISION_LOG2);
		int world_z_acceleration_fp8 = (x_acceleration_fp8 * m[0 * mr + 2] + y_acceleration_fp8 * m[1 * mr + 2] + z_acceleration_fp8 * m[2 * mr + 2]) >> (Camera.PRECISION_LOG2);
		world_z_acceleration_fp8 += GRAVITY_ACCELERATION_FP8;
		// speed = curr_speed + acceleration * dt. dt = 1;
		mSpeedX_fp8 += world_x_acceleration_fp8;
		mSpeedY_fp8 += world_y_acceleration_fp8;
		mSpeedZ_fp8 += world_z_acceleration_fp8;
		//
		mCG_X_fp8 += mSpeedX_fp8;
		mCG_X = mCG_X_fp8 >> 8;
		mCG_Y_fp8 += mSpeedY_fp8;
		mCG_Y = mCG_Y_fp8 >> 8;
		mCG_Z_fp8 += mSpeedZ_fp8;
		mCG_Z = mCG_Z_fp8 >> 8;
		//--------------------------------------------------------------------------
		mFrontLeftWheelBottom = WHEEL_BOTTOM_Z +
			Math.min( Math.max( mRoadToFrontLeftWheelDistance_fp >> Camera.PRECISION_LOG2, 0 ), SPRINT_TRAVEL );

		mFrontRightWheelBottom = WHEEL_BOTTOM_Z +
			Math.min( Math.max( mRoadToFrontRightWheelDistance_fp >> Camera.PRECISION_LOG2, 0 ), SPRINT_TRAVEL );

		mRearLeftWheelBottom = WHEEL_BOTTOM_Z +
			Math.min( Math.max( mRoadToRearLeftWheelDistance_fp >> Camera.PRECISION_LOG2, 0 ), SPRINT_TRAVEL );

		mRearRightWheelBottom = WHEEL_BOTTOM_Z +
			Math.min( Math.max( mRoadToRearRightWheelDistance_fp >> Camera.PRECISION_LOG2, 0 ), SPRINT_TRAVEL );
		//--------------------------------------------------------------------------
		setZangle( z_angle );//+ mRotationSpeedZ );// z-1 - left
		//
		mRotationSpeedY += y_rotation_acceleration;
		mRotationSpeedX += x_rotation_acceleration;
		int angle = x_angle + mRotationSpeedX;
		if( angle > 65 * Camera.ANGLE_360 / 360 ) {
			angle = 65 * Camera.ANGLE_360 / 360;
			mRotationSpeedX = 0;
		} else if( angle < -65 * Camera.ANGLE_360 / 360 ) {
			angle = -65 * Camera.ANGLE_360 / 360;
			mRotationSpeedX = 0;
		}
		setXangle( angle );// x-1 - forward UP
		//
		angle = y_angle + mRotationSpeedY;
		if( angle > 65 * Camera.ANGLE_360 / 360 ) {
			angle = 65 * Camera.ANGLE_360 / 360;
			mRotationSpeedY = 0;
		} else if( angle < -65 * Camera.ANGLE_360 / 360 ) {
			angle = -65 * Camera.ANGLE_360 / 360;
			mRotationSpeedY = 0;
		}
		setYangle( angle );// y-1 - clockwise
	}
}