class Section implements Comparable<Section> {
	/** camera.mZ * m[0 * mr + 2] */
	private static int sDz0 = 0;
	/** camera.mZ * m[1 * mr + 2] */
	private static int sDz1 = 0;
	/** camera.mZ * m[2 * mr + 2] */
	private static int sDz2 = 0;
	//
	/** to check start line and is odd or is even */
	final int mCoordIndex;
	final Coord3D mLeft;
	final Coord3D mRight;
	/** to access to the piece colors */
	final TrackPiece mPiece;
	Section mNextSection;
	final Coord3D mLeftTop = new Coord3D();
	final Coord3D mLeftBottom = new Coord3D();
	final Coord3D mRightTop = new Coord3D();
	final Coord3D mRightBottom = new Coord3D();
	int mLeftTopX = 0;
	int mLeftTopY = 0;
	int mLeftBottomX = 0;
	int mLeftBottomY = 0;
	int mRightTopX = 0;
	int mRightTopY = 0;
	int mRightBottomX = 0;
	int mRightBottomY = 0;
	/** square of view distance */
	int mViewDistance_d64 = 0;
	private boolean mIsNeedCalculateScreenCoords = true;	
	//
	Section(int coordIndex, Coord3D left, Coord3D right, TrackPiece piece, Section nextSection) {
		mLeft = left;
		mRight = right;
		mNextSection = nextSection;
		mPiece = piece;
		mCoordIndex = coordIndex;
	}
	/** Calculate general values before calculate coordinates */
	static void prepareToCalculate(final Camera camera) {
		final int mr = 3;
		final int[] m = camera.mMatrix;
		sDz0 = camera.mZ * m[0 * mr + 2];
		sDz1 = camera.mZ * m[1 * mr + 2];
		sDz2 = camera.mZ * m[2 * mr + 2];
	}
	/** Before use call the prepareToCalculate function */
	void calculateCoordinates(final Camera camera) {
		int dx, dy, dz, temp;
		final int mr = 3;
		// to speed
		final int[] m = camera.mMatrix;
		//
		mIsNeedCalculateScreenCoords = true;
		//
		dx = mLeft.mX - camera.mX;
		dy = mLeft.mY - camera.mY;
		dz = mLeft.mZ - camera.mZ;
		//
		temp = dx * m[0 * mr + 0] + dy * m[0 * mr + 1];
		mLeftTop.mX = temp + dz * m[0 * mr + 2];
		mLeftBottom.mX = temp - sDz0;
		temp = dx * m[1 * mr + 0] + dy * m[1 * mr + 1];
		mLeftTop.mY = temp + dz * m[1 * mr + 2];
		mLeftBottom.mY = temp - sDz1;
		temp = dx * m[2 * mr + 0] + dy * m[2 * mr + 1];
		mLeftTop.mZ = temp + dz * m[2 * mr + 2];
		mLeftBottom.mZ = temp - sDz2;
		//
		dx = mRight.mX - camera.mX;
		dy = mRight.mY - camera.mY;
		dz = mRight.mZ - camera.mZ;
		//
		temp = dx * m[0 * mr + 0] + dy * m[0 * mr + 1];
		mRightTop.mX = temp + dz * m[0 * mr + 2];
		mRightBottom.mX = temp - sDz0;
		temp = dx * m[1 * mr + 0] + dy * m[1 * mr + 1];
		mRightTop.mY = temp + dz * m[1 * mr + 2];
		mRightBottom.mY = temp - sDz1;
		temp = dx * m[2 * mr + 0] + dy * m[2 * mr + 1];
		mRightTop.mZ = temp + dz * m[2 * mr + 2];
		mRightBottom.mZ = temp - sDz2;
		//
		dx = mLeftTop.mX >> (Camera.PRECISION_LOG2 + 3);// +3 to prevent overflow
		dy = mLeftTop.mY >> (Camera.PRECISION_LOG2 + 3);
		temp = dx * dx + dy * dy;
		if( dy < 0 ) temp = -temp;
		mViewDistance_d64 = temp;
	}
	void worldToScreen() {
		if( mIsNeedCalculateScreenCoords ) {
			mIsNeedCalculateScreenCoords = false;
			int scale = (mLeftTop.mY + (Render.CAMERA_FOV << Camera.PRECISION_LOG2)) / Render.SCALE;
			mLeftTopX = GameView.SCREEN_HALFWIDTH + (mLeftTop.mX / scale);
			mLeftTopY = Render.LOOK_Y - (mLeftTop.mZ / scale);
	
			scale = (mLeftBottom.mY + (Render.CAMERA_FOV << Camera.PRECISION_LOG2)) / Render.SCALE;
			mLeftBottomX = GameView.SCREEN_HALFWIDTH + (mLeftBottom.mX / scale);
			mLeftBottomY = Render.LOOK_Y - (mLeftBottom.mZ / scale);
	
			scale = (mRightTop.mY + (Render.CAMERA_FOV << Camera.PRECISION_LOG2)) / Render.SCALE;
			mRightTopX = GameView.SCREEN_HALFWIDTH + (mRightTop.mX / scale);
			mRightTopY = Render.LOOK_Y - (mRightTop.mZ / scale);
	
			scale = (mRightBottom.mY + (Render.CAMERA_FOV << Camera.PRECISION_LOG2)) / Render.SCALE;
			mRightBottomX = GameView.SCREEN_HALFWIDTH + (mRightBottom.mX / scale);
			mRightBottomY = Render.LOOK_Y - (mRightBottom.mZ / scale);
			//
			if( Render.sIsPreview ) {
				mLeftBottomX = mLeftTopX;
				mRightBottomX = mRightTopX;
			}
		}
	}
	@Override
	public int compareTo(Section s) {
		return mViewDistance_d64 - s.mViewDistance_d64;
	}
}