import java.awt.Color;
/** Track's data. Maximal track diameter must be less then 32768.
 * This limitation due horizon distance equals 65536
 */
abstract class Track {
	static final int TRACKS_CENTER_X = 16384;
	static final int TRACKS_CENTER_Y = 16384;
	/** track's name */
	char[] mName;
	/** color for start line on top side */
	Color mRoadStartLineColor;
	/** color for dark side lines on top */
	Color mRoadLineColorDark;
	/** color for light side lines on top */
	Color mRoadLineColorLight;
	/** Road width to interpolate a height */
	final int mRoadWidth = 384;
	/** Scenery for the track */
	Model[] mScenery = SceneryData.mMixed;
	/** Segment count, not implemented */
	//int mNumTrackSegments;
	/** Number of track coordinates */
	int mNumTrackCoords;
	/** Amount of nitrous oxide for standard league, sec */
	int mStandardBoost;
	/** Amount of nitrous oxide for super league, sec */
	//int mSuperBoost;
	int mBestLapTimeMs10;
	final char[] mBestLapPilot = new char[12];
	int mBestRaceTimeMs10;
	final char[] mBestRacePilot = new char[12];
	/** 3d coordinates for left and right top corners */
	TrackPiece[] mPieces;
	int mSpeedDeviation_fp8;// int8
	int mMaxSpeed_fp8;

	ITrackAnimation mAnimation;

	Track() {
		mRoadStartLineColor = new Color( 0xff, 0xff, 0xff );
		mRoadLineColorDark = new Color( 0x40, 0x00, 0x00 );
		mRoadLineColorLight = new Color( 0xe0, 0xe0, 0x00 );
		resetBestTime();
		clear();
	}
	final void resetBestTime() {
		mBestLapTimeMs10 = 0x00ffffff;// < 7fffffff / 10
		mBestRaceTimeMs10 = 0x00ffffff;// < 7fffffff / 10
		for( int i = 0; i < mBestLapPilot.length; i++ ) mBestLapPilot[i] = '-';
		for( int i = 0; i < mBestRacePilot.length; i++ ) mBestRacePilot[i] = '-';
	}
	static void setName(final char[] target, final char[] name) {
		int i = 0;
		for( ; i < name.length; i++ ) {
			target[i] = name[i];
		}
		for( ; i < target.length; i++ ) {
			target[i] = ' ';
		}
	}
	final void clear() {
		mPieces = null;
	}
	/** calculate count of coordinates */
	final void updateData() {
		int coords_count = 0;
		for( int i = 0; i < mPieces.length; i++ ) {
			final TrackPiece tp = mPieces[i];
			coords_count += tp.mCoords.length;
			coords_count -= 2;
			// km/h to px/frame
			tp.mRecommendedSpeed_fp8 = tp.mRecommendedSpeed_fp8 * (256 * (int)GameView.REFRESH_PERIOD * 2 / 36);
		}
		mNumTrackCoords = coords_count;
		// km/h to px/frame
		mMaxSpeed_fp8 = mMaxSpeed_fp8 * (256 * (int)GameView.REFRESH_PERIOD * 2 / 36);
		mSpeedDeviation_fp8 = mSpeedDeviation_fp8 * (256 * (int)GameView.REFRESH_PERIOD * 2 / 36);
	}
	abstract void init();
	/** scale down track's coordinates */
	/*final Track scaleDown(int xscale, int yscale, int zscale) {
		for( int n = 0; n < mPieces.length; n++ ) {
			final Coord3D[] coords = mPieces[n].mCoords;
			for( int i = 0; i < coords.length; i++ ) {
				coords[i].mX /= xscale;
				coords[i].mY /= yscale;
				coords[i].mZ /= zscale;
			}
		}
		return this;
	}*/
	/** scale up track's coordinates */
	/*final Track scaleUp(int xscale, int yscale, int zscale) {
		for( int n = 0; n < mPieces.length; n++ ) {
			final Coord3D[] coords = mPieces[n].mCoords;
			for( int i = 0; i < coords.length; i++ ) {
				coords[i].mX *= xscale;
				coords[i].mY *= yscale;
				coords[i].mZ *= zscale;
			}
		}
		return this;
	}*/
	/** shift track's coordinates */
	/*final Track shift(int dx, int dy, int dz) {
		for( int n = 0; n < mPieces.length; n++ ) {
			final Coord3D[] coords = mPieces[n].mCoords;
			for( int i = 0; i < coords.length; i++ ) {
				coords[i].mX += dx;
				coords[i].mY += dy;
				coords[i].mZ += dz;
			}
		}
		return this;
	}*/
	/** to center XY track's coordinates around 0.0 */
	/*final Track centerXYToZero() {
		int xmin = 32767;
		int xmax = 0;
		int ymin = 32767;
		int ymax = 0;
		for( int n = 0; n < mPieces.length; n++ ) {
			final Coord3D[] coords = mPieces[n].mCoords;
			for( int i = 0; i < coords.length; i++ ) {
				if( xmin > coords[i].mX ) xmin = coords[i].mX;
				if( xmax < coords[i].mX ) xmax = coords[i].mX;
				if( ymin > coords[i].mY ) ymin = coords[i].mY;
				if( ymax < coords[i].mY ) ymax = coords[i].mY;
			}
		}
		int midx = (xmax - xmin) >> 1;
		int midy = (ymax - ymin) >> 1;
		for( int n = 0; n < mPieces.length; n++ ) {
			final Coord3D[] coords = mPieces[n].mCoords;
			for( int i = 0; i < coords.length; i++ ) {
				coords[i].mX -= midx;
				coords[i].mY -= midy;
			}
		}
		return this;
	}*/
	static final boolean isRoadSegmentBefore(final TrackPiece piece, int offset, int x, int y) {
		final Coord3D[] c = piece.mCoords;
		final Coord3D sc = c[offset];
		final Coord3D ec = c[offset + 1];
		// orientation = (end.x - start.x) * (y - start.y) - (end.y - start.y) * (x - start.x)
		// orientation > 0 : left
		// orientation < 0 : right
		// orientation == 0 : on
		return ( (ec.mX - sc.mX) * (y - sc.mY) + (sc.mY - ec.mY) * (x - sc.mX) < 0 );
	}
	static final boolean isRoadSegmentBehind(final TrackPiece piece, int offset, int x, int y) {
		// orientation = (end.x - start.x) * (y - start.y) - (end.y - start.y) * (x - start.x)
		// orientation > 0 : left
		// orientation < 0 : right
		// orientation == 0 : on
		final Coord3D[] c = piece.mCoords;
		final Coord3D sc = c[offset];
		final Coord3D ec = c[offset + 1];
		return ( (ec.mX - sc.mX) * (y - sc.mY) + (sc.mY - ec.mY) * (x - sc.mX) > 0 );
	}
	static final boolean isOnRoad(final TrackPiece piece, int offset, int x, int y) {
		// orientation = (end.x - start.x) * (y - start.y) - (end.y - start.y) * (x - start.x)
		// orientation > 0 : left
		// orientation < 0 : right
		// orientation == 0 : on
		int next_offset = offset + 2;
		final Coord3D[] c = piece.mCoords;
		final Coord3D left_near = c[offset];
		final Coord3D right_near = c[offset + 1];
		final Coord3D left_far = c[next_offset];
		final Coord3D right_far = c[next_offset + 1];
		//
		if( (left_far.mX - left_near.mX) * (y - left_near.mY) -
			(left_far.mY - left_near.mY) * (x - left_near.mX) <= 0 ) {
			if( (right_far.mX - right_near.mX) * (y - right_near.mY) -
				(right_far.mY - right_near.mY) * (x - right_near.mX) >= 0 )
				return true;
		}
		return false;
	}
	/** Convert cos^2 to angle from 0 to 90 degrees */
	private static final int COS_PRECISION_LOG2 = 8;
	/** Cosines to angle from 0 to 90 degrees */
	private static final int[] sCosToAngle;
	//private static final int[] sCos2ToInvCos;
	static {
		sCosToAngle = new int[(1 << COS_PRECISION_LOG2) + 1];
		for( int i = 0; i < sCosToAngle.length; i++ ) {
			final double cos = Math.sqrt( ((double)i) / (1 << COS_PRECISION_LOG2) );
			final int angle = (int)(Camera.ANGLE_180 * Math.acos( cos ) / Math.PI);
			sCosToAngle[i] = angle;
		}
		/*
		sCos2ToInvCos = new int[(1 << PRECISION_LOG2) + 1];
		for( int i = 0; i < sCos2ToInvCos.length; i++ ) {
			final double cos = Math.sqrt( ((double)i) / (1 << PRECISION_LOG2) );
			sCos2ToInvCos[i] = (1 << PRECISION_LOG2) - (int)(cos * (1 << PRECISION_LOG2));
		}
		*/
	}
	/**
	 * 
	 * @param piece
	 * @param offset
	 * @return angle from -180 degrees to +180 degrees
	 */
	static int getSectionAngle(final TrackPiece piece, int offset) {
		final Coord3D[] c = piece.mCoords;
		final Coord3D near_left = c[offset++];// c[offset];
		final Coord3D near_right = c[offset++];// c[offset + 1];
		final Coord3D far_left = c[offset++];// c[next_offset];
		final Coord3D far_right = c[offset];// c[next_offset + 1];
		//
		final int dx = ((far_left.mX + far_right.mX) - (near_left.mX + near_right.mX)) >> 1;
		final int dy = ((far_left.mY + far_right.mY) - (near_left.mY + near_right.mY)) >> 1;
		// cos^2 = (x1 * x2 + y1 * y2)/((x1 * x1 + y1 * y1) * (x2 * x2 + y2 * y2))
		// for base vector x1 = 0, y1 = 1, so:
		// cos^2 = (y2 * y2)/(x2 * x2 + y2 * y2)
		final int dy2 = dy * dy;
		final int cos2 = (dy2 << COS_PRECISION_LOG2) / (dx * dx + dy2);
		int angle = sCosToAngle[Math.abs( cos2 )];
		if( dy < 0 ) angle = Camera.ANGLE_180 - angle;
		if( dx < 0 ) return -angle;
		return angle;
	}
	/*
	final int getSectionSlopeCos(int offset) {
		int next_offset = offset + 2;
		if( next_offset >= mNumTrackCoords ) next_offset = 0;
		//
		final Coord3D near_left = mCoords[offset];
		final Coord3D near_right = mCoords[offset + 1];
		final Coord3D far_left = mCoords[next_offset];
		final Coord3D far_right = mCoords[next_offset + 1];
		//
		final int dx = ((far_left.x + far_right.x) - (near_left.x + near_right.x)) >> 1;
		final int dy = ((far_left.y + far_right.y) - (near_left.y + near_right.y)) >> 1;
		final int dz = ((far_left.z + far_right.z) - (near_left.z + near_right.z)) >> 1;
		// cos^2 = (x1 * x2 + y1 * y2)/((x1 * x1 + y1 * y1) * (x2 * x2 + y2 * y2))
		// for base vector x1 = 0, y1 = 1, so:
		// cos^2 = (y2 * y2)/(x2 * x2 + y2 * y2)
		final int p = (dx * dx + dy * dy);
		final int cos2 = (p << PRECISION_LOG2) / (p + dz * dz);
		return sCos2ToInvCos[cos2] << (GameView.PRECISION_LOG2 - PRECISION_LOG2);
	}*/
}